#!/usr/bin/env python
import sys,os, argparse, warnings, subprocess
from collections import defaultdict
from pandas.errors import EmptyDataError
import pyfastx
from tqdm import tqdm
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    open_file_reader,
    # open_file_writer,
    read_pickle, 
    # write_pickle,
    read_json,
    # write_json,
    # build_logger,
    # reset_logger,
    get_timestamp,
    # format_duration,
    # format_header,
    format_bytes,
    get_file_size,
    profile_peak_memory,
    RunShellCommand,
)

@profile_peak_memory
def process_genomic_databases_and_check_inputs(fasta, feature_mapping, genomes, logger, config):  
   # Fasta
    genes_from_fasta = set()
    for id, seq in tqdm(pyfastx.Fasta(fasta, build_index=False), f"Loading fasta: {fasta}"):
        genes_from_fasta.add(id)
    
    # Feature mapping
    # ===============
    gene_to_data = defaultdict(dict)
    genome_to_data = defaultdict(dict)
    features_from_data = set()
    with open_file_reader(feature_mapping) as f:
        first_line = f.readline().strip()
        assert "\t" in first_line
        fields = first_line.split("\t")
        assert len(fields) in {3,4}, "Expecting the following fields (No header): [id_gene, features, id_genome, (Optional: id_genome_cluster)]"
        if len(fields) == 3:
            id_gene, features, id_genome  = fields
            gene_to_data[id_gene]["features"] = eval(features)
            gene_to_data[id_gene]["id_genome"] = id_genome
            gene_to_data[id_gene]["id_genome_cluster"] = None
            genomes_from_feature_mapping.add(id_genome)
            
            for line in tqdm(f, f"Loading feature mapping: {feature_mapping}"):
                line = line.strip()
                if line:
                    id_gene, features, id_genome = line.split("\t")
                    gene_to_data[id_gene]["features"] = eval(features)
                    gene_to_data[id_gene]["id_genome"] = id_genome
                    gene_to_data[id_gene]["id_genome_cluster"] = None
                    genome_to_data[id_genome][["id_genome_cluster"]] = None
                    
        elif len(fields) == 4:
            id_gene, features, id_genome, id_genome_cluster  = fields
            gene_to_data[id_gene]["features"] = eval(features)
            gene_to_data[id_gene]["id_genome"] = id_genome
            gene_to_data[id_gene]["id_genome_cluster"] = id_genome_cluster
            genome_to_data[id_genome]["id_genome_cluster"] = id_genome_cluster

            for line in tqdm(f, f"Loading feature mapping with genome clusters: {feature_mapping}"):
                line = line.strip()
                if line:
                    id_gene, features, id_genome, id_genome_cluster = line.split("\t")
                    gene_to_data[id_gene]["features"] = eval(features)
                    gene_to_data[id_gene]["id_genome"] = id_genome
                    gene_to_data[id_gene]["id_genome_cluster"] = id_genome_cluster
                    genome_to_data[id_genome]["id_genome_cluster"] = id_genome_cluster
                    
        config["contains_genome_cluster_mapping"] = True
        
    for id_gene, data in gene_to_data.items():
        features_from_data.update(data["features"])

    # Check overlap of genes between --fasta and --feature_mapping
    genes_from_feature_mapping = set(gene_to_data.keys())
    genomes_from_feature_mapping = set(genome_to_data.keys())
    if genes_from_fasta != genes_from_feature_mapping:
        A_exclusive = genes_from_fasta - genomes_from_feature_mapping
        B_exclusive = genomes_from_feature_mapping - genes_from_fasta
        msg = "--fasta must contain same genes in --feature_mapping"
        if A_exclusive:
            msg += f"\nN={len(A_exclusive)} genes in --fasta that are not in --feature_mapping"
        if B_exclusive:
            msg += f"\nN={len(B_exclusive)} genes in --feature_mapping that are not in --fasta"
        logger.critical(msg)
        raise IndexError(msg)
    config["number_of_genes"] = len(genes_from_feature_mapping)
    config["number_of_features"] = len(features_from_data)
    
    # Check if features as KEGG orthologs
    config["feature_type_is_kegg_ortholog"] = True
    for id_feature in features_from_data:
        conditions = [
            id_feature.startswith("K"),
            len(id_feature) == 6,
            id_feature[1:].isnumeric(),
        ]
        if not all(conditions):
            config["feature_type_is_kegg_ortholog"] = False
    
    # Genomes
    # =======
    genomes_with_filepaths = set()
    if genomes is not None:
        with open_file_reader(genomes) as f:
            for line in tqdm(f, f"Loading genomes: {genomes}"):
                line = line.strip()
                if line:
                    id_genome, filepath = line.split("\t")
                    genome_to_data[id_genome]["filepath"] = filepath
                    genomes_with_filepaths.add(id_genome)
                    
        if genomes_from_feature_mapping != genomes_with_filepaths:
            A_exclusive = genomes_from_feature_mapping - genomes_with_filepaths
            B_exclusive = genomes_with_filepaths - genomes_from_feature_mapping
            msg = "--feature_mapping and --genomes genome sets are different"
            if A_exclusive:
                msg += f"\nN={len(A_exclusive)} genomes in --feature_mapping that are not in --genomes"
            if B_exclusive:
                msg += f"\nN={len(B_exclusive)} genomes in --genomes that are not in --feature_mapping"
            logger.warn(msg)
            # raise IndexError(msg)
        
        config["contains_genome_filepaths"] = True
        
    config["number_of_genomes"] = len(genomes_with_filepaths)
    config["timestamp"] = get_timestamp()

    return config, gene_to_data, genome_to_data

# Load genome database and check inputs
def update_genome_database_with_fasta_filepaths_and_check_inputs(index_directory, genomes, logger, config):
    # Read database files
    gene_to_data = read_pickle(os.path.join(index_directory, "database", "gene_to_data.pkl.gz"))
    genome_to_data = read_pickle(os.path.join(index_directory, "database", "genome_to_data.pkl.gz"))
    genomes_from_feature_mapping = set(genome_to_data.keys())

    # Genomes
    genomes_with_filepaths = set()
    with open_file_reader(genomes) as f:
        for line in tqdm(f, f"Loading genomes: {genomes}"):
            line = line.strip()
            if line:
                id_genome, filepath = line.split("\t")
                genome_to_data[id_genome]["filepath"] = filepath
                genomes_with_filepaths.add(id_genome)
                
    if genomes_from_feature_mapping != genomes_with_filepaths:
        A_exclusive = genomes_from_feature_mapping - genomes_with_filepaths
        B_exclusive = genomes_with_filepaths - genomes_from_feature_mapping
        msg = "--feature_mapping and --genomes genome sets are different"
        if A_exclusive:
            msg += f"\nN={len(A_exclusive)} genomes in --feature_mapping that are not in --genomes"
        if B_exclusive:
            msg += f"\nN={len(B_exclusive)} genomes in --genomes that are not in --feature_mapping"
        logger.warn(msg)
        # raise IndexError(msg)
    
    config["contains_genome_filepaths"] = True
    config["timestamp"] = get_timestamp()

    return config, gene_to_data, genome_to_data

# Check pathway database
def load_pathway_database_and_check_inputs(index_directory, gene_to_data, logger, config):
    # Read database files
    pathway_to_data = read_pickle(os.path.join(index_directory, "database", "pathway_to_data.pkl.gz"))

    # Features from gene identifier mapping
    features_from_data = set()
    for id_gene, data in gene_to_data.items():
        features_from_data.update(data["features"])
        
    # Features from pathways
    features_from_pathways = set()
    for id_pathway, data in pathway_to_data.items():
        features_from_pathways.update(set(data["ko_to_nodes"].keys()))
        
    # Overlapping features
    overlapping_features = features_from_data & features_from_pathways
    number_of_overlapping_features = len(overlapping_features)
                
    if number_of_overlapping_features > 0:
        A_exclusive = features_from_data - features_from_pathways
        B_exclusive = features_from_pathways - features_from_data
        logger.info(f"Number of features from data: {len(features_from_data)} [Exclusive: {len(A_exclusive)}]")
        logger.info(f"Number of features from pathways: {len(features_from_pathways)} [Exclusive: {len(B_exclusive)}]")
        logger.info(f"Number of features overlapping between data and pathways: {number_of_overlapping_features}")
        config["number_of_features_in_pathways"] = len(features_from_pathways)
        config["number_of_features_overlapping_in_pathways"] = number_of_overlapping_features

    else:
        msg = f"No features {len(features_from_data)} from gene_to_data.pkl.gz overlap with the pathway_to_data.pkl.gz database.\nTo address this, either 1) remove pathway database, 2) confirm pathway database features match gene feature types, or 3) use KEGG orthologs for gene feature types"
        msg += "\n\nHere are a few of the features from gene_to_data.pkl.gz: {}".format("\n\t".join(list(features_from_data)[:min(5, len(features_from_data))]))
        msg += "\n\nHere are a few of the features from pathway_to_data.pkl.gz: {}".format("\n\t".join(list(features_from_data)[:min(5, len(features_from_pathways))]))
        logger.critical(msg)
        raise IndexError(msg)
    
    config["timestamp"] = get_timestamp()

    
    return config, pathway_to_data

# Check salmon index files
def check_salmon_index(salmon_index_directory, logger):
    expected_files = [
        'seq.bin', 'info.json', 'pre_indexing.log', 'ref_indexing.log', 'ctable.bin', 'refAccumLengths.bin', 
        'mphf.bin', 'versionInfo.json', 'duplicate_clusters.tsv', 'ctg_offsets.bin', 'reflengths.bin', 
        'pos.bin', 'refseq.bin', 'complete_ref_lens.bin', 'rank.bin',
        ]
    
    missing_files = list()
    empty_files = list()
    if os.path.exists(salmon_index_directory):
        for filename in expected_files:
            filepath = os.path.join(salmon_index_directory, filename)
            if not os.path.exists(filepath):
                missing_files.append(filepath)
                
            filesize = get_file_size(filepath)
            logger.info(f"[salmon_index] {filepath} ({format_bytes(filesize)})")
            if filesize < 1:
                empty_files.append(filepath)
        if any(missing_files):
            msg = f"Salmon index directory is corrupted.  The following files are missing: {missing_files}]"
            logger.critical(msg)
            raise FileNotFoundError(msg)
        if any(empty_files):
            msg = f"Salmon index directory is corrupted.  The following files are empty: {empty_files}]"
            logger.critical(msg)
            raise EmptyDataError(msg)
    else:
        msg = f"Salmon index directory does not exist: {salmon_index_directory} therefore updates are not applicable.  Please rerun with --fasta, --feature_mapping, and --genomes arguments"
        logger.critical(msg)
        raise FileNotFoundError(msg)

# Run Salmon (salmon index --keepDuplicates --threads ${N_JOBS} --transcripts ${FASTA} --index ${INDEX})
def run_salmon_indexer(logger, log_directory, salmon_executable, n_jobs, fasta, index_directory, index_options):    
    cmd = RunShellCommand(
        command=[
            salmon_executable,
            "index",
            "--keepDuplicates",
            "--threads",
            n_jobs,
            "--transcripts",
            fasta,
            "--index",
            os.path.join(index_directory, "salmon_index"),
            index_options,   
            ], 
        name="salmon_indexer",
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd

# Run Sylph (sylph sketch -t ${N_JOBS} --gl ${GENOMES} -o ${INDEX} -k ${K} --min-spacing ${MIN_SPACING})
def run_sylph_genomes_sketcher(logger, log_directory, sylph_executable, n_jobs, genome_filepaths, index_directory, k, minimum_spacing, subsampling_rate, sylph_sketch_options):
    cmd = RunShellCommand(
        command=[
            sylph_executable,
            "sketch",
            "-t",
            n_jobs,
            "--gl",
            genome_filepaths,
            "-o",
            os.path.join(index_directory, "database", "genomes"),
            "-k",
            k,
            "--min-spacing",
            minimum_spacing,
            "-c",
            subsampling_rate,
            sylph_sketch_options,
        ],
        name="sylph_genomes_sketcher",
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd

    
# Download KEGG Pathway Database
def run_kegg_pathway_downloader(logger, log_directory, pathway_database_downloader_executable,  index_directory, no_intermediate_files):
    cmd = RunShellCommand(
        command=[
            pathway_database_downloader_executable,
            "--download",
            "--force",
            "--no_intermediate_files" if no_intermediate_files else "",
            "--database",
            os.path.join(index_directory, "database", "pathway_to_data.pkl.gz"),
    
        ],
        name="kegg_pathway_downloader",
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd


#!/usr/bin/env python
import sys,os, argparse, warnings, subprocess, glob
from collections import defaultdict
import pandas as pd
# from pandas.errors import EmptyDataError
# from Bio.SeqIO.FastaIO import SimpleFastaParser
from tqdm import tqdm
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    # open_file_reader,
    # open_file_writer,
    read_pickle, 
    # write_pickle,
    read_json,
    # write_json,
    build_logger,
    # reset_logger,
    # format_duration,
    # format_header,
    get_file_size,
    format_bytes,
    # get_directory_tree,
    # get_directory_size,
    # get_md5hash_from_directory,
    check_file,
    RunShellCommand,
)
from leviathan.index import(
    check_salmon_index,
)

from leviathan.profile_pathway import(
    run_salmon_quant,
    reformat_gene_abundance,
    reformat_feature_abundance,
    build_wide_feature_prevalence_matrix,
    build_feature_prevalence_dictionary,
    build_feature_pathway_dictionary,
    calculate_pathway_coverage,
    aggregate_pathway_abundance_and_append_coverage,
    aggregate_feature_abundance_for_clusters,
)

def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} -1 forward.fq[.gz] -2 reverse.fq[.gz] -n sample_name -o project_directory --index_directory path/to/leviathan_index/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser_io = parser.add_argument_group('I/O arguments')
    parser_io.add_argument("-1","--forward_reads", type=str,  help = "path/to/forward_reads.fq[.gz] (Cannot be used with -s/--read_sketch)")
    parser_io.add_argument("-2","--reverse_reads", type=str,  help = "path/to/reverse_reads.fq[.gz] (Cannot be used with -s/--read_sketch)")
    # parser_io.add_argument("--ont", type=str,  help = "path/to/ont_reads.fq[.gz] (Cannot be used with -s/--read_sketch)")
    parser_io.add_argument("-n", "--name", type=str, required=True, help="Name of sample")
    parser_io.add_argument("-o","--project_directory", type=str, default="leviathan_output/profiling/pathway", help = "path/to/project_directory (e.g., leviathan_output/profiling/pathway]")
    parser_io.add_argument("-d","--index_directory", type=str, required=True, help = "path/to/index_directory/")
    parser_io.add_argument("-f","--output_format", type=str, choices={"tsv", "parquet"}, default="parquet", help = "Output format [Default: parquet]")

    # Utilities
    parser_utility = parser.add_argument_group('Utility arguments')
    parser_utility.add_argument("-p","--n_jobs", type=int, default=1,  help = "Number of threads to use.  Use -1 for all available. [Default: 1]")

    # Salmon
    parser_salmon_quant = parser.add_argument_group('salmon quant arguments')
    parser_salmon_quant.add_argument("--salmon_executable", type=str, help="salmon executable [Default: $PATH]")
    parser_salmon_quant.add_argument("-m", "--minimum_score_fraction", type=float, default=0.87, help="The fraction of the optimal possible alignment score that a mapping must achieve in order to be considered \"valid\" --- should be in (0,1]. (relaxed: 0.65, strict: 0.87) [Default: 0.87]")
    parser_salmon_quant.add_argument("--salmon_include_mappings", action="store_true", help="salmon quant| Include mappings")
    parser_salmon_quant.add_argument("--salmon_gzip", action="store_true", help="salmon quant | Gzip quant.sf")

    parser_salmon_quant.add_argument("--salmon_quant_options", type=str, default="", help="salmon quant| More options (e.g. --arg=1 ) https://salmon.readthedocs.io/en/latest/ [Default: '']")

    # Samtools
    parser_samtools = parser.add_argument_group('samtools arguments')
    parser_samtools.add_argument("--samtools_executable", type=str, help="samtools executable [Default: $PATH]")
    parser_samtools.add_argument("--alignment_format", type=str, choices={"sam", "bam", "sorted.bam"}, default="sorted.bam", help="samtools alignment format [Default: sorted.bam]")

    # Features
    parser_features = parser.add_argument_group('Features arguments')
    parser_features.add_argument("--no_split_feature_abundances", action="store_true", help="Do not split feature abundances [Default: if one gene is assigned to N enzyme features, then distribute the abundance equally across all N features)")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan profile-pathway")

    # Commands
    logger.info(f"Command: {sys.argv}")

    # Threads
    if opts.n_jobs == -1:
        from multiprocessing import cpu_count 
        opts.n_jobs = cpu_count()
        logger.info(f"Setting --n_jobs to maximum threads {opts.n_jobs}")

    assert opts.n_jobs >= 1, "--n_jobs must be ≥ 1.  To select all available threads, use -1."
    
    # Executables
    # * Salmon
    if not opts.salmon_executable:
        opts.salmon_executable = os.path.join(bin_directory, "salmon")
    if not os.path.exists(opts.salmon_executable):
        msg = f"salmon executable not doesn't exist: {opts.salmon_executable}"
        logger.critical(msg)
        raise FileNotFoundError(msg)

    # * Samtools
    if not opts.samtools_executable:
        opts.samtools_executable = os.path.join(bin_directory, "samtools")
    if not os.path.exists(opts.samtools_executable):
        msg = f"samtools executable not doesn't exist: {opts.samtools_executable}"
        logger.critical(msg)
        raise FileNotFoundError(msg)
    
    # Config
    config = read_json(os.path.join(opts.index_directory, "config.json"))

    # Check Salmon database
    logger.info("Checking Salmon index") 
    check_salmon_index(
        salmon_index_directory=os.path.join(opts.index_directory, "salmon_index"),
        logger=logger,
        )
    
    # Check gene metadata
    gene_data_filepath = os.path.join(opts.index_directory, "database", "gene_to_data.pkl.gz")
    logger.info(f"Checking gene metadata: {gene_data_filepath}")
    check_file(gene_data_filepath, minimum_filesize=48)
    
    # Check gene metadata
    genome_data_filepath = os.path.join(opts.index_directory, "database", "genome_to_data.pkl.gz")
    logger.info(f"Checking genome metadata: {genome_data_filepath}")
    check_file(genome_data_filepath, minimum_filesize=48)
    
    # Check pathway metadata
    if config["contains_pathways"]:
        pathway_data_filepath = os.path.join(opts.index_directory, "database", "pathway_to_data.pkl.gz")
        logger.info(f"Checking pathway metadata: {pathway_data_filepath}")
        check_file(pathway_data_filepath, minimum_filesize=48)
    else:
        logger.warning(f"No pathways available in index: {opts.index_directory}.  Feature-level profiling will be completed but pathway abundances and coverage will not be in output.")

    # Output
    output_directory = os.path.join(opts.project_directory, opts.name)
    os.makedirs(output_directory, exist_ok=True)
    os.makedirs(os.path.join(output_directory, "output"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "intermediate"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "logs"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "tmp"), exist_ok=True)
    
    # =====================
    # Loading gene metadata
    # =====================
    logger.info("Loading gene data") 
    gene_to_data = read_pickle(os.path.join(opts.index_directory, "database", "gene_to_data.pkl.gz"))

    # =======================
    # Loading genome metadata
    # =======================
    logger.info("Loading genome data") 
    genome_to_data = read_pickle(os.path.join(opts.index_directory, "database", "genome_to_data.pkl.gz"))

    # =======================
    # Loading pathway metadata
    # =======================
    if config["contains_pathways"]:
        logger.info("Loading pathway data") 
        pathway_to_data = read_pickle(os.path.join(opts.index_directory, "database", "pathway_to_data.pkl.gz"))

    # ====================
    # Build Salmon profiler
    # ====================
    # Run Salmon quant
    logger.info("Running Salmon quant")
    cmd_salmon_quant = run_salmon_quant(
        logger=logger,
        log_directory=os.path.join(output_directory, "logs"), 
        salmon_executable=opts.salmon_executable, 
        samtools_executable=opts.samtools_executable,
        n_jobs=opts.n_jobs, 
        output_directory=os.path.join(output_directory, "intermediate"), 
        index_directory=opts.index_directory,
        forward_reads=opts.forward_reads, 
        reverse_reads=opts.reverse_reads, 
        minimum_score_fraction=opts.minimum_score_fraction, 
        include_mappings=opts.salmon_include_mappings,
        alignment_format=opts.alignment_format,
        salmon_gzip=opts.salmon_gzip,
        salmon_quant_options=opts.salmon_quant_options, 
    )
       
    # ===============================
    # Abundance and prevalence tables
    # ===============================
    level="genome"
    
    # Gene abundances
    gene_abundance_filepath = os.path.join(output_directory, "output", f"gene_abundances.{level}s.{opts.output_format}")
    if opts.output_format != "parquet":
        gene_abundance_filepath += ".gz"
    logger.info(f"[level={level}] Reformatting gene abundance: {gene_abundance_filepath}")
    
    filepath_quantsf = os.path.join(output_directory, "intermediate", "quant.sf")
    if opts.salmon_gzip:
        filepath_quantsf += ".gz"
    df_quant = pd.read_csv(filepath_quantsf, sep="\t", index_col=0)
    df_gene_abundance = reformat_gene_abundance(df_quant, gene_to_data)
    if opts.output_format == "parquet":
        df_gene_abundance.to_parquet(gene_abundance_filepath, index=True)
    elif opts.output_format == "tsv":
        df_gene_abundance.to_csv(gene_abundance_filepath, sep="\t")
        
    # Feature abundances
    feature_abundance_filepath = os.path.join(output_directory, "output", f"feature_abundances.{level}s.{opts.output_format}")
    if opts.output_format != "parquet":
        feature_abundance_filepath += ".gz"
    logger.info(f"[level={level}] Calculating feature abundance: {feature_abundance_filepath}")
    df_feature_abundance = reformat_feature_abundance(df_gene_abundance, gene_to_data, split_feature_abundances=not opts.no_split_feature_abundances)
    if opts.output_format == "parquet":
        df_feature_abundance.to_parquet(feature_abundance_filepath, index=True)
    elif opts.output_format == "tsv":
        df_feature_abundance.to_csv(feature_abundance_filepath, sep="\t")
        
    # Feature prevalence
    feature_prevalence_filepath = os.path.join(output_directory, "output", f"feature_prevalence.{level}s.{opts.output_format}")
    if opts.output_format != "parquet":
        feature_prevalence_filepath += ".gz"
    logger.info(f"[level={level}] Calculating feature prevalence: {feature_prevalence_filepath}")
    df_feature_prevalence = build_wide_feature_prevalence_matrix(df_feature_abundance, threshold=0)
    if opts.output_format == "parquet":
        df_feature_prevalence.to_parquet(feature_prevalence_filepath, index=True)
    elif opts.output_format == "tsv":
        df_feature_prevalence.to_csv(feature_prevalence_filepath, sep="\t")
        
    # Feature prevalence (binary)
    feature_prevalence_binary_filepath = os.path.join(output_directory, "output", f"feature_prevalence-binary.{level}s.{opts.output_format}")
    if opts.output_format != "parquet":
        feature_prevalence_binary_filepath += ".gz"
    logger.info(f"[level={level}] Reformatting binary feature prevalence: {feature_prevalence_binary_filepath}")
    df_feature_prevalence_binary = (df_feature_prevalence > 0).astype(int)
    if opts.output_format == "parquet":
        df_feature_prevalence_binary.to_parquet(feature_prevalence_binary_filepath, index=True)
    elif opts.output_format == "tsv":
        df_feature_prevalence_binary.to_csv(feature_prevalence_binary_filepath, sep="\t")
        
    if config["contains_pathways"]:
        logger.info(f"[level={level}] Building binary feature prevalence to dictionary")
        genome_to_features = build_feature_prevalence_dictionary(df_feature_prevalence_binary)
        logger.info(f"[level={level}] Building feature to pathways dictionary")
        feature_to_pathways = build_feature_pathway_dictionary(pathway_to_data)
        logger.info(f"[level={level}] Calculating pathway coverage")
        coverages = calculate_pathway_coverage(genome_to_features, pathway_to_data)
        
        pathway_abundances_filepath = os.path.join(output_directory, "output", f"pathway_abundances.{level}s.{opts.output_format}")
        if opts.output_format != "parquet":
            pathway_abundances_filepath += ".gz"
        logger.info(f"[level={level}] Aggregating pathway abundances and appending coverages: {pathway_abundances_filepath}")
        df_pathway_abundances = aggregate_pathway_abundance_and_append_coverage(df_feature_abundance, feature_to_pathways, coverages, index_names = [f"id_{level}", "id_pathway"])
        if opts.output_format == "parquet":
            df_pathway_abundances.to_parquet(pathway_abundances_filepath, index=True)
        elif opts.output_format == "tsv":
            df_pathway_abundances.to_csv(pathway_abundances_filepath, sep="\t")

    if config["contains_genome_cluster_mapping"]:
        level="genome_cluster"
        
        # Feature abundance
        feature_abundance_filepath = os.path.join(output_directory, "output", f"feature_abundances.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            feature_abundance_filepath += ".gz"
        logger.info(f"[level=genome_cluster] Calculating feature abundance: {feature_abundance_filepath}")
        df_feature_abundance = aggregate_feature_abundance_for_clusters(df_feature_abundance, genome_to_data)
        if opts.output_format == "parquet":
            df_feature_abundance.to_parquet(feature_abundance_filepath, index=True)
        elif opts.output_format == "tsv":
            df_feature_abundance.to_csv(feature_abundance_filepath, sep="\t")
        
        # Feature ratio
        feature_ratio_filepath = os.path.join(output_directory, "output", f"feature_prevalence-ratio.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            feature_ratio_filepath += ".gz"
        logger.info(f"[level=genome_cluster] Calculating feature prevalence ratios: {feature_ratio_filepath}")
        df_feature_prevalence_ratio = (df_feature_prevalence > 0).groupby(lambda x: genome_to_data[x]["id_genome_cluster"]).mean()
        if opts.output_format == "parquet":
            df_feature_prevalence_ratio.to_parquet(feature_ratio_filepath, index=True)
        elif opts.output_format == "tsv":
            df_feature_prevalence_ratio.to_csv(feature_ratio_filepath, sep="\t")
        
        # Feature prevalence
        feature_prevalence_filepath = os.path.join(output_directory, "output", f"feature_prevalence.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            feature_prevalence_filepath += ".gz"
        logger.info(f"[level=genome_cluster] Calculating feature prevalence: {feature_prevalence_filepath}")
        df_feature_prevalence = df_feature_prevalence.groupby(lambda x: genome_to_data[x]["id_genome_cluster"]).sum()        
        if opts.output_format == "parquet":
            df_feature_prevalence.to_parquet(feature_prevalence_filepath, index=True)
        elif opts.output_format == "tsv":
            df_feature_prevalence.to_csv(feature_prevalence_filepath, sep="\t")
        
        # Feature prevalence (binary)
        feature_prevalence_binary_filepath = os.path.join(output_directory, "output", f"feature_prevalence-binary.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            feature_prevalence_binary_filepath += ".gz"
        logger.info(f"[level=genome_cluster] Reformatting binary feature prevalence: {feature_prevalence_binary_filepath}")
        df_feature_prevalence_binary = (df_feature_prevalence > 0).astype(int)
        
        if opts.output_format == "parquet":
            df_feature_prevalence_binary.to_parquet(feature_prevalence_binary_filepath, index=True)
        elif opts.output_format == "tsv":
            df_feature_prevalence_binary.to_csv(feature_prevalence_binary_filepath, sep="\t")
        
        if config["contains_pathways"]:
            logger.info(f"[level={level}] Building binary feature prevalence to dictionary")
            genome_to_features = build_feature_prevalence_dictionary(df_feature_prevalence_binary)
            logger.info(f"[level={level}] Building feature to pathways dictionary")
            feature_to_pathways = build_feature_pathway_dictionary(pathway_to_data)
            logger.info(f"[level={level}] Calculating pathway coverage")
            coverages = calculate_pathway_coverage(genome_to_features, pathway_to_data)
            
            pathway_abundances_filepath = os.path.join(output_directory, "output", f"pathway_abundances.{level}s.{opts.output_format}")
            if opts.output_format != "parquet":
                pathway_abundances_filepath += ".gz"
            logger.info(f"[level={level}] Aggregating pathway abundances and appending coverages: {pathway_abundances_filepath}")
            df_pathway_abundances = aggregate_pathway_abundance_and_append_coverage(df_feature_abundance, feature_to_pathways, coverages, index_names = [f"id_{level}", "id_pathway"])
            
            if opts.output_format == "parquet":
                df_pathway_abundances.to_parquet(pathway_abundances_filepath, index=True)
            elif opts.output_format == "tsv":
                df_pathway_abundances.to_csv(pathway_abundances_filepath, sep="\t")


    # ========
    # Complete
    # ========    
    logger.info(f"Completed pathway profiling: {opts.name}")
    for filepath in glob.glob(os.path.join(output_directory, "output","*")):
        filesize = get_file_size(filepath, format=True)
        logger.info(f"Output: {filepath} ({filesize})")

    logger.info(f"Completed running leviathan-profile-pathway for {opts.name}: {opts.project_directory}")

if __name__ == "__main__":
    main()
    
    

    



#!/usr/bin/env python
import sys
import os
import argparse

from pyexeggutor import (
    read_json,
    build_logger,
    format_bytes,
    get_directory_size,
    get_md5hash_from_file,
)

__program__ = os.path.split(sys.argv[0])[-1]


def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} --fasta path/to/cds.fasta --feature_mapping path/to/features.tsv --genomes path/to/genomes.tsv  --index_directory path/to/leviathan_index/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-n", "--number_of_rows", type=int, default=10, help="Number of rows to show for parquet [Default: 10]")
    parser.add_argument("-e", "--xarray_engine", type=str, choices={"h5netcdf", "netcdf4"}, default="h5netcdf", help = "Xarray backend engine [Default: h5netcdf]")
    group = parser.add_argument_group('Input Options (choose one)')
    mutex_group = group.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument("-d", "--index_directory", type=str, help="path/to/index_directory/")
    mutex_group.add_argument("-p", "--parquet", type=str, help="path/to/table.parquet")
    mutex_group.add_argument("-x", "--netcdf", type=str, help="path/to/dataset.nc")
    
    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan info")

    # Commands
    logger.info(f"Command: {sys.argv}")
     
    if opts.index_directory:
        logger.info(f"Providing info for index: {opts.index_directory}")
        print()
        # Size
        size_in_bytes = get_directory_size(opts.index_directory)
        logger.info(f"Database size: {format_bytes(size_in_bytes)} ({size_in_bytes} bytes)")

        # Config
        path_config = os.path.join(opts.index_directory, "config.json")
        config = read_json(path_config)
        logger.info(f"Config: {path_config}")
        for k, v in config.items():
            logger.info(f"  |--- {k}: {v}")

        # Hash
        path_md5 = os.path.join(opts.index_directory, "md5hashes.json")
        hashes = read_json(path_md5)
        logger.info(f"MD5 hashes: {path_md5}")
        for k, v in hashes.items():
            logger.info(f"  |--- {k}: {v}")
        
    elif opts.parquet:
        import fastparquet as fp
        logger.info(f"Providing preview of parquet: {opts.parquet}")
        pf = fp.ParquetFile(opts.parquet)
        df = pf.head(opts.number_of_rows)
        print()
        df.to_csv(sys.stdout, sep="\t")
        
    elif opts.netcdf:
        import xarray as xr
        logger.info(f"Providing preview of Xarray NetCDF: {opts.netcdf}")
        ds = xr.open_dataset(opts.netcdf, engine=opts.xarray_engine)
        print()
        print(ds)

    logger.info(f"Completed running leviathan-info: {opts.index_directory}")

        
if __name__ == "__main__":
    main()
    
    

    



#!/usr/bin/env python
import sys
import os
import argparse
from collections import defaultdict
from itertools import (
    product, 
    chain,
)
from pandas.errors import EmptyDataError
from tqdm import tqdm
import xarray as xr

from pyexeggutor import (
    build_logger,
)

from leviathan.profile_taxonomy import (
    merge_taxonomic_profiling_tables_as_xarray,
)
from leviathan.profile_pathway import (
    merge_pathway_profiling_tables_as_xarray,
)

__program__ = os.path.split(sys.argv[0])[-1]


def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} -t path/to/taxonomic_profiling_directory/ -p path/to/pathway_profiling_directory/ -o path/to/output_directory/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser.add_argument("-t","--taxonomic_profiling_directory", type=str, help = "path/to/profiling/taxonomy/")
    parser.add_argument("-p","--pathway_profiling_directory", type=str, help = "path/to/profiling/pathway/")
    parser.add_argument("-o","--output_directory", type=str,  help = "path/to/output_directory. Default is either --taxonomic_profiling_directory and --pathway_profiling_directory")
    parser.add_argument("-z","--fillna_with_zeros", action="store_true", help = "Fill missing values with 0.  This will take a lot longer to write to disk.")
    parser.add_argument("-e", "--xarray_engine", type=str, choices={"h5netcdf", "netcdf4"}, default="h5netcdf", help = "Xarray backend engine [Default: h5netcdf]")
    parser.add_argument("-c", "--xarray_compression_level", type=int, choices=set(range(0, 10)), default=4, help = "netCDF gzip compression level. Use 0 for no compression. [Default: 4]")
    
    # ------------------
    # Pending options
    # ------------------

    # ------------------
    # Deprecated options
    # ------------------
    # parser.add_argument("-f","--output_format", type=str, choices={"tsv", "pickle", "parquet"}, default="parquet", help = "Output format [Default: parquet]")
    # parser.add_argument("-s","--sparse", action="store_true", help = "Output pd.SparseDtype.  This will take a lot longer to write to disk.  Only applicable when --output_format=pickle.")
    # parser.add_argument("--no_transpose_pathway_profiling", action="store_true", help = "Do not transpose pathway profiling tables.  If you do not transpose them, it will use much more memory to read/write unless --output_format pickle")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan merge")

    # Commands
    logger.info(f"Command: {sys.argv}")
     
    # I/O
    # if opts.output_format == "parquet":
        # logger.warn(f"--output_format parquet results in transposed output relative to tsv and pickle (n=genomes, m=features).  To avoid memory constraints, parquet will have features as rows and genomes/genome-clusters as columns.")
        
    ## Taxonomic Profiling
    proceed_with_merging_taxonomic_profiles = False
    taxonomic_profiling_output_directory = None
    if opts.taxonomic_profiling_directory:
        if os.path.exists(opts.taxonomic_profiling_directory):
            proceed_with_merging_taxonomic_profiles = True
            if not opts.output_directory:
                taxonomic_profiling_output_directory = opts.taxonomic_profiling_directory
            else:
                taxonomic_profiling_output_directory = opts.output_directory
            logger.info(f"Creating taxonomic profiling output directory (if it does not exist): {taxonomic_profiling_output_directory}")
            os.makedirs(taxonomic_profiling_output_directory, exist_ok=True)
        else:
            msg = f"Taxonomic profiling output directory does not exist: {opts.taxonomic_profiling_directory}"
            logger.critical(msg)
            raise Exception(msg)

    ## Pathway Profiling
    proceed_with_merging_pathway_profiles = False
    pathway_profiling_output_directory = None
    if opts.pathway_profiling_directory:
        if os.path.exists(opts.pathway_profiling_directory):
            proceed_with_merging_pathway_profiles = True
            if not opts.output_directory:
                pathway_profiling_output_directory = opts.pathway_profiling_directory
            else:
                pathway_profiling_output_directory = opts.output_directory
            logger.info(f"Creating pathway profiling output directory (if it does not exist): {pathway_profiling_output_directory}")
            os.makedirs(pathway_profiling_output_directory, exist_ok=True)
        else:
            msg = f"Pathway profiling output directory does not exist: {opts.pathway_profiling_directory}"
            logger.critical(msg)
            raise Exception(msg)
        
    # Run
    ## Taxonomic Profiling
    if proceed_with_merging_taxonomic_profiles:
        for level in ["genomes", "genome_clusters"]:
            logger.info(f"Merging taxonomic profiles for level={level}")

            try:
                # Filepath
                filepath = os.path.join(taxonomic_profiling_output_directory, f"taxonomic_abundances.{level}.nc")

                X = merge_taxonomic_profiling_tables_as_xarray(
                    profiling_directory=opts.taxonomic_profiling_directory, 
                    level=level, 
                    fillna_with_zeros=bool(opts.fillna_with_zeros), 
                )
                n = X.sizes["samples"]
                m = X.sizes[level]

                error_msg = f"Merging taxonomic profiles for level={level} in {opts.taxonomic_profiling_directory} resulted in empty xr.DataArray"
                info_msg = f"Taxonomy profiles for level={level} have {n} samples, {m} {level}"

                if len(X) == 0:
                    raise EmptyDataError(error_msg)
                    
                logger.info(info_msg)

                if opts.xarray_compression_level:
                    logger.info(f"Setting gzip compression: {opts.xarray_compression_level}")
                    X.encoding.update({"compression": "gzip", "compression_opts": opts.xarray_compression_level})
                
                # logger.info(f"Adding DataArray ({name}) to DataSet group {group}")
                # group_to_dataset[group][name] = X
                mode = "w"
                logger.info(f"Writing output: {filepath} [mode={mode}]")
                X.to_netcdf(filepath, engine=opts.xarray_engine, mode=mode)
                del X
     
            except Exception as e:
                logger.warning(f"No level={level} files found in {opts.taxonomic_profiling_directory}: {e}")
                
        logger.info(f"Completed merging taxonomic profiling tables: {taxonomic_profiling_output_directory}")

    ## Pathway Profiling
    if proceed_with_merging_pathway_profiles:

        levels = ["genomes", "genome_clusters"]
        abundance_data_types = ["feature_abundances", "pathway_abundances"] #  "gene_abundances",
        prevalence_data_types = ["feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio"]
        metrics = ["number_of_reads", "tpm", "coverage"]
        
        argument_combinations = chain(
            product(levels, abundance_data_types, metrics),
            product(levels, prevalence_data_types, ["number_of_reads"]), # Expects an argument but it's not actually used
        )
        
        # Specify group file mode
        group_to_filemode = dict()
        group_to_dataset = defaultdict(xr.Dataset)
        for level, data_type, metric in argument_combinations:
            illegal_conditions = [
                # (level == "genome_cluster") and (data_type == "gene_abundances"),
                (level == "genomes") and (data_type == "feature_prevalence-ratio"),
                (data_type != "pathway_abundances") and (metric == "coverage"),
            ]
            if not any(illegal_conditions):
                try:
                    # Group
                    group = (
                        data_type.split("_")[0], 
                        level,
                    )
                    
                    # Mode
                    if group not in group_to_filemode:
                        group_to_filemode[group] = "w"
                    else:
                        group_to_filemode[group] = "a"
                    mode = group_to_filemode[group]
                    
                    # Filepath
                    filepath = os.path.join(pathway_profiling_output_directory, f"{group[0]}.{group[1]}.nc")

                    # Variable name
                    name = metric if "abundances" in data_type else data_type.split("_")[-1]
                    
                    # Merge pathway profiling
                    X = merge_pathway_profiling_tables_as_xarray(
                        profiling_directory=opts.pathway_profiling_directory, 
                        data_type=data_type, 
                        level=level, 
                        metric=metric, 
                        fillna_with_zeros=bool(opts.fillna_with_zeros), 
                        )

                    if data_type in prevalence_data_types:
                        error_msg = f"Merging pathway profiles for level={level}, data_type={data_type} in {opts.pathway_profiling_directory} resulted in empty xr.DataArray"
                        info_msg = f"Pathway profiles for level={level}, data_type={data_type} have {X.shape[0]} samples, {X.shape[1]} {level}, and {X.shape[2]} features"

                    else:
                        error_msg = f"Merging pathway profiles for level={level}, data_type={data_type}, metric={metric} in {opts.pathway_profiling_directory} resulted in empty xr.DataArray"
                        info_msg = f"Pathway profiles for level={level}, data_type={data_type}, metric={metric} have {X.shape[0]} samples, {X.shape[1]} {level}, and {X.shape[2]} features"

                    if len(X) == 0:
                        raise EmptyDataError(error_msg)
                    
                    logger.info(info_msg)
                    if opts.xarray_compression_level:
                        logger.info(f"Setting gzip compression: {opts.xarray_compression_level}")
                        X.encoding.update({"compression": "gzip", "compression_opts": opts.xarray_compression_level})
                    
                    logger.info(f"Adding DataArray ({name}) to DataSet group {group}")
                    group_to_dataset[group][name] = X
                    logger.info(f"Writing output: {filepath} [mode={mode}]")
                    group_to_dataset[group].to_netcdf(filepath, engine=opts.xarray_engine, mode=mode)
                    del X

                except Exception as e:
                    if data_type in prevalence_data_types:
                        logger.warning(f"Not able to merge {data_type}.{level} files from {opts.pathway_profiling_directory}: {e}")
                    else:
                        logger.warning(f"Not able to merge {data_type}.{level}.{metric} files from {opts.pathway_profiling_directory}: {e}")
                    
        logger.info(f"Completed merging pathway profiling tables: {pathway_profiling_output_directory}")
    logger.info(f"Completed running leviathan-merge: {opts.output_directory}")


if __name__ == "__main__":
    main()


    



#!/usr/bin/env python
import sys, os, glob
from tqdm import tqdm
from collections import defaultdict
import pandas as pd
from pyexeggutor import (
    open_file_reader,
    check_argument_choice,
)

# Annotations
def read_annotations(path:str, format="pykofamsearch"):
    
    """
    Reads feature annotations from a given file and returns a dictionary mapping each gene to its associated features.

    Parameters
    ----------
    path : str
        Path to the annotation file
    format : str
        Format of the annotation file. Options: pykofamsearch, pykofamsearch-reformatted, pyhmmsearch, pyhmmsearch-reformatted, veba-pfam, veba-kofam, veba-cazy, veba-uniref, veba-mibig, veba-vfdb, veba-amr, custom

    Returns
    -------
    gene_to_features : dict
        Mapping of gene identifiers to their associated features
    """
    check_argument_choice(
        query=format, 
        choices={"pykofamsearch", "pykofamsearch-reformatted", "pyhmmsearch","pyhmmsearch-reformatted", "veba-pfam","veba-kofam","veba-cazy","veba-uniref", "veba-mibig", "veba-vfdb","veba-amr", "custom"},
        )
    
    if format in {"pykofamsearch", "pyhmmsearch", "custom", "pykofamsearch-reformatted", "pyhmmsearch-reformatted"}:
        f_annotations = open_file_reader(path)

        gene_to_features = defaultdict(set)    
        if format != "custom":
            next(f_annotations)
            
        if format in {"pykofamsearch-reformatted", "pyhmmsearch-reformatted"}:
            for line in tqdm(f_annotations, desc="Extracting feature annotations"):
                line = line.strip()
                if line:
                    id_gene, number_of_hits, features, *extra = line.split("\t")
                    gene_to_features[id_gene] = set(eval(features))

        for line in tqdm(f_annotations, desc="Extracting feature annotations"):
            line = line.strip()
            if line:
                id_gene, id_feature, *extra = line.split("\t")
                gene_to_features[id_gene].add(id_feature)
        f_annotations.close()



    else:
        df_annotations = pd.read_csv(path, sep="\t", index_col=0, header=[0,1])
        
        # HMM-based annotations
        if format in {"veba-pfam", "veba-kofam", "veba-amr"}:
            column = {
                "veba-pfam":("Pfam", "ids"),
                "veba-kofam":("KOfam", "ids"),
                "veba-amr":("NCBIfam-AMR", "ids"),
            }[format]
            gene_to_features = df_annotations[column].map(lambda x: set(eval(x)))
        
        # Diamond-based annotations
        else:
            column = {
                "veba-cazy":("CAZy", "sseqid"),
                "veba-uniref":("UniRef", "sseqid"),
                "veba-mibig":("MIBiG", "sseqid"),
                "veba-vfdb":("VFDB", "sseqid"),
            }[format]
            
            gene_to_features = defaultdict(set)
            for id_gene, id_feature in df_annotations[column].items():
                gene_to_features[id_gene].add(id_feature)
    return gene_to_features


                




#!/usr/bin/env python
import sys,os, argparse, warnings, subprocess, shutil
from collections import defaultdict
from importlib.resources import files as resource_files
# from pandas.errors import EmptyDataError
# from Bio.SeqIO.FastaIO import SimpleFastaParser
from tqdm import tqdm
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    open_file_reader,
    open_file_writer,
    read_pickle, 
    write_pickle,
    read_json,
    write_json,
    build_logger,
    reset_logger,
    format_duration,
    format_header,
    format_bytes,
    get_timestamp,
    get_directory_tree,
    get_directory_size,
    get_md5hash_from_directory,
    RunShellCommand,
)

from leviathan.index import(
    process_genomic_databases_and_check_inputs,
    update_genome_database_with_fasta_filepaths_and_check_inputs,
    load_pathway_database_and_check_inputs,
    check_salmon_index,
    run_salmon_indexer,
    run_sylph_genomes_sketcher,
    run_kegg_pathway_downloader,
)

DEFAULT_PATHWAY_DATABASE = str(resource_files('kegg_pathway_profiler').joinpath('data/database.pkl.gz'))
if not os.path.exists(DEFAULT_PATHWAY_DATABASE):
    DEFAULT_PATHWAY_DATABASE = None

def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} --fasta path/to/cds.fasta --feature_mapping path/to/features.tsv --genomes path/to/genomes.tsv  --index_directory path/to/leviathan_index/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser_io = parser.add_argument_group('I/O arguments')
    parser_io.add_argument("-f","--fasta", type=str,  help = "path/to/cds.fasta") # default="stdin"?
    parser_io.add_argument("-m","--feature_mapping", type=str,  help = "path/to/feature_mapping.tsv [id_gene, feature_set, id_genome, (Optional: id_genome_cluster)] (No header)")
    parser_io.add_argument("-g","--genomes", type=str, help = "path/to/genomes.tsv [id_genome, path/to/genome] (No header)")
    parser_io.add_argument("-d","--index_directory", type=str, required=True, help = "path/to/index_directory/ (Recommended: leviathan_output/index/ if this will only be used for one project or a centralized location if it will be used for multiple projects)")
    parser_io.add_argument("-u", "--update_with_genomes", action="store_true",  help = "Update databases with genomes for Sylph sketches")

    # Utilities
    parser_utility = parser.add_argument_group('Utility arguments')
    parser_utility.add_argument("-p","--n_jobs", type=int, default=1,  help = "Number of threads to use.  Use -1 for all available. [Default: 1]")
    parser_utility.add_argument("--no_check", action="store_true",  help = "Don't check to ensure --genomes set equals --feature_mapping genome set")

    # Salmon
    parser_salmon_index = parser.add_argument_group('salmon index arguments')
    parser_salmon_index.add_argument("--salmon_executable", type=str, help="salmon executable [Default: $PATH]")
    parser_salmon_index.add_argument("--salmon_index_options", type=str, default="", help="salmon index | More options (e.g. --arg=1 ) https://salmon.readthedocs.io/en/latest/ [Default: '']")

    # Sylph
    parser_sylph_sketch = parser.add_argument_group('Sylph genomes sketcher arguments (Fastq)')
    parser_sylph_sketch.add_argument("--sylph_executable", type=str, help="Sylph executable [Default: $PATH]")
    parser_sylph_sketch.add_argument("--sylph_k", type=int, choices={21,31}, default=31,  help="Sylph |  Value of k. Only k = 21, 31 are currently supported. [Default: 31]")
    parser_sylph_sketch.add_argument("--sylph_minimum_spacing", type=int,  default=30,  help="Sylph |  Minimum spacing between selected k-mers on the genomes [Default: 30]")
    parser_sylph_sketch.add_argument("--sylph_subsampling_rate", type=int, default=200,  help="Sylph | Subsampling rate.	[Default: 200]")
    parser_sylph_sketch.add_argument("--sylph_sketch_options", type=str, default="", help="Sylph | More options for `sylph sketch` (e.g. --arg=1 ) [Default: '']")

    # Pathway
    parser_pathways = parser.add_argument_group('Pathway database arguments')
    parser_pathways.add_argument("--pathway_database_downloader_executable", type=str, help="KEGG Pathway profiler `build-pathway-database.py` executable. Ignored if --database is provided. [Default: $PATH]")
    parser_pathways.add_argument("--no_intermediate_files", action="store_true",  help = "Don't write intermediate files when downloading database")
    parser_pathways.add_argument("--pathway_database", type=str, default=DEFAULT_PATHWAY_DATABASE, help=f"Pathway database formatted as a Python pkl[.gz].  See documentation for details.  If no database is provided, then a database will be generated if KEGG orthologs are provided as features. [Default: {DEFAULT_PATHWAY_DATABASE}]")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan index")

    # Commands
    logger.info(f"Command: {sys.argv}")

    # Checks
    if not opts.no_check:
        genomes_with_filepaths = set()
        with open_file_reader(opts.genomes) as f:
            for line in f:
                line = line.strip()
                if line:
                    id_genome, filepath = line.split("\t")
                    genomes_with_filepaths.add(id_genome)

        genomes_with_features = set()
        with open_file_reader(opts.feature_mapping) as f:
            for line in f:
                line = line.strip()
                if line:
                    fields = line.split("\t")
                    id_genome = fields[2]
                    genomes_with_features.add(id_genome)
        if genomes_with_filepaths == genomes_with_features:
            msg = "All genomes in --genomes are in --feature_mapping"
            logger.info(msg)
        else:
            msg = f"{len(genomes_with_filepaths) - len(genomes_with_features)} genomes in --genomes are missing from --feature_mapping"
            logger.critical(msg)
            parser.error(msg)
     
    # Post-processing argument dependencies
    if opts.update_with_genomes:
        if not opts.genomes:
            msg = "--genomes is required when --update_with_genomes is specified"
            logger.critical(msg)
            parser.error(msg)
        # Check salmon index
        check_salmon_index(os.path.join(opts.index_directory, "salmon_index"))
        
        # Required files without --update_with_genomes
        if not opts.fasta or not opts.feature_mapping:
            msg = "--fasta and --feature_mapping are required when --update_with_genomes is not specified"
            logger.critical(msg)
            parser.error(msg)
        if not opts.genomes:
            logger.warning("--genomes not provided but can incoporated post hoc by rerunning with --update_with_genomes")

    # Threads
    if opts.n_jobs == -1:
        from multiprocessing import cpu_count 
        opts.n_jobs = cpu_count()
        logger.info(f"Setting --n_jobs to maximum threads {opts.n_jobs}")

    assert opts.n_jobs >= 1, "--n_jobs must be ≥ 1.  To select all available threads, use -1."
    
    # Executables
    # * Salmon
    if not opts.salmon_executable:
        opts.salmon_executable = os.path.join(bin_directory, "salmon")
    if not os.path.exists(opts.salmon_executable):
        msg = f"salmon executable doesn't exist: {opts.salmon_executable}"
        logger.critical(msg)
        raise FileNotFoundError(msg)
    
    # * Sylph
    if not opts.sylph_executable:
        opts.sylph_executable = os.path.join(bin_directory, "sylph")
    if not os.path.exists(opts.sylph_executable):
        msg = f"sylph executable doesn't exist: {opts.sylph_executable}"
        logger.critical(msg)
        raise FileNotFoundError(msg)
    
    # * KEGG Pathway Profiler
    if opts.pathway_database:
        logger.info(f"Database provided: {opts.pathway_database}")
        if opts.pathway_database_downloader_executable:
            logger.info(f"Ignoring --pathway_database_downloader_executable {opts.pathway_database_downloader_executable}")
    else:
        if not opts.pathway_database_downloader_executable:
            opts.pathway_database_downloader_executable = os.path.join(bin_directory, "build-pathway-database.py")
        if not os.path.exists(opts.pathway_database_downloader_executable):
            msg = f"KEGG Pathway Profiler's `build-pathway-database.py` executable doesn't exist: {opts.pathway_database_downloader_executable}. Please install or provide a custom database."
            logger.critical(msg)
            raise FileNotFoundError(msg)
    
    # Index directory
    if all([
        os.path.exists(opts.index_directory),
        not opts.update_with_genomes,
        ]):
        msg = f"--index_directory {opts.index_directory} already exists.  If you want to update with genomes, please use --update_with_genomes or remove directory to overwrite"
        logger.critical(msg)
        raise FileExistsError(msg)
    
    os.makedirs(os.path.join(opts.index_directory, "database"), exist_ok=True)
    os.makedirs(os.path.join(opts.index_directory, "logs"), exist_ok=True)
    os.makedirs(os.path.join(opts.index_directory, "tmp"), exist_ok=True)

    if not opts.update_with_genomes:
        # Setup config
        logger.info("Setting up config")

        config = dict(
            fasta_filepath=opts.fasta,
            feature_mapping_filepath=opts.feature_mapping,
            contains_genome_cluster_mapping=False,
            contains_genome_filepaths=False,
            timestamp=get_timestamp(),
        )
        # Genomes and genes database
        # --------------------------
        # Process genomes/genes and check inputs
        logger.info("Processing and checking input files")
        
        config, gene_to_data, genome_to_data = process_genomic_databases_and_check_inputs(
            fasta=opts.fasta, 
            feature_mapping=opts.feature_mapping, 
            genomes=opts.genomes,
            logger=logger,
            config=config,
            )
        
        # Pathway database
        # ----------------
        # Fetch database
        config["contains_pathways"] = False
        if opts.pathway_database:
            logger.info(f"Pathway database provided and copying to index: {opts.pathway_database}")
            shutil.copy(opts.pathway_database, os.path.join(opts.index_directory, "database", "pathway_to_data.pkl.gz"), follow_symlinks=True)
            config["contains_pathways"] = True
        else:
            if config["feature_type_is_kegg_ortholog"]:
                # Run KEGG Pathway Downloader
                logger.info("Running KEGG Pathway downloader")
                cmd_kegg_pathway_downloader = run_kegg_pathway_downloader(
                    logger=logger,
                    log_directory = os.path.join(opts.index_directory, "logs"),
                    pathway_database_downloader_executable=opts.pathway_database_downloader_executable,
                    index_directory=opts.index_directory,
                    no_intermediate_files=bool(opts.no_intermediate_files),
                    )
                config["contains_pathways"] = True
            else:
                logger.warning("No pathway database was provided and features were not determined to be KEGG orthologs so automatic database compilation is not available.  Pathway profiling will not include pathway coverage or abundances.")

        # Check pathway database
        if config["contains_pathways"]:
            config, pathway_to_data = load_pathway_database_and_check_inputs(
                index_directory=opts.index_directory, 
                gene_to_data=gene_to_data, 
                logger=logger, 
                config=config,
                )

        # Write database files
        logger.info("Writing config and database files")
        write_json(config, os.path.join(opts.index_directory,  "config.json"))
        write_pickle(gene_to_data, os.path.join(opts.index_directory, "database", "gene_to_data.pkl.gz"))
        write_pickle(genome_to_data, os.path.join(opts.index_directory, "database", "genome_to_data.pkl.gz"))

        # ==================
        # Build Salmon Index
        # ==================
        cmd_salmon_indexer = run_salmon_indexer(
                        logger=logger,
                        log_directory = os.path.join(opts.index_directory, "logs"),
                        salmon_executable=opts.salmon_executable,
                        n_jobs=opts.n_jobs,
                        fasta=opts.fasta,
                        index_directory=opts.index_directory,
                        index_options=opts.salmon_index_options,
        )
        
    else:
        # Load config
        config_filepath = os.path.join(opts.index_directory, "config.json")
        logger.info(f"Loading config: {config_filepath}")
        config = read_json(config_filepath)

        # Process and check inputs
        logger.info("Loading previously built database and checking genomes for database update")
        config, gene_to_data, genome_to_data = update_genome_database_with_fasta_filepaths_and_check_inputs(
            index_directory=opts.index_directory, 
            genomes=opts.genomes,
            logger=logger,
            config=config,
            )
        
        # Update database files
        genome_to_data_filepath = os.path.join(opts.index_directory, "database", "genome_to_data.pkl.gz")
        logger.info(f"Rewriting config and database files: {genome_to_data_filepath}")
        write_json(config, config_filepath)
        write_pickle(genome_to_data, genome_to_data_filepath)
        

    # ==================
    # Build Sylph Sketch
    # ==================
    if config["contains_genome_filepaths"]:
        logger.info("Writing genome filepaths for Sylph")

        # Write filepaths
        genome_filepaths = os.path.join(opts.index_directory, "tmp", "genome_filepaths.list")
        with open_file_writer(genome_filepaths) as f:
            for id_genome, data in tqdm(genome_to_data.items(), f"Writing filepaths: {genome_filepaths}"):
                print(data["filepath"], file=f)
                
        # Run Sylph
        logger.info("Running Sylph genome sketcher")
        cmd_sylph_genomes_sketcher = run_sylph_genomes_sketcher(
            logger=logger,
            log_directory = os.path.join(opts.index_directory, "logs"),
            sylph_executable=opts.sylph_executable, 
            n_jobs=opts.n_jobs, 
            genome_filepaths=genome_filepaths, 
            index_directory=opts.index_directory,
            k=opts.sylph_k, 
            minimum_spacing=opts.sylph_minimum_spacing, 
            subsampling_rate=opts.sylph_subsampling_rate, 
            sylph_sketch_options=opts.sylph_sketch_options,
        )

    else:
        logger.warning("--genomes not provided so Leviathan is not building Sylph sketches")
        
    # ========
    # Hash
    # ========   
    md5hash_filepath = os.path.join(opts.index_directory, "md5hashes.json")
    logger.info(f"Calculating md5 hashes: {md5hash_filepath}")
    file_to_hash = get_md5hash_from_directory(opts.index_directory)
    write_json(file_to_hash, md5hash_filepath)

    # ========
    # Complete
    # ========    
    logger.info(f"Completed building leviathan index: {opts.index_directory}")
    logger.info(f"Directory size of leviathan index: {format_bytes(get_directory_size(opts.index_directory))}")
    logger.info(f"Directory structure of leviathan index:\n{get_directory_tree(opts.index_directory, ascii=True)}")
    logger.info(f"Completed running leviathan-index: {opts.index_directory}")

if __name__ == "__main__":
    main()
    
    

    



#!/usr/bin/env python
import sys,os, argparse, warnings, subprocess, shutil
from collections import defaultdict
# from importlib.resources import files as resource_files
# from pandas.errors import EmptyDataError
import pandas as pd
from tqdm import tqdm
import pyfastx
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    open_file_reader,
    open_file_writer,
    # read_pickle, 
    # write_pickle,
    # read_json,
    write_json,
    build_logger,
    # reset_logger,
    format_duration,
    # format_header,
    format_bytes,
    # get_timestamp,
    get_directory_tree,
    get_directory_size,
    get_md5hash_from_directory,
    # RunShellCommand,
)
from leviathan.utils import (
    read_annotations,
)


def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} -i/--input path/to/input.tsv -a path/to/annotations.tsv -o/--output_directory path/to/output_directory/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser.add_argument("-i","--input", type=str, default="stdin", help = "path/to/input.tsv[.gz] [id_genome, path/to/assembly.fasta, path/to/cds.fasta. [Optional: id_genome_cluster]] (No header, Tab delimited)")
    parser.add_argument("-a","--annotations", type=str, help = "path/to/annotations.tsv[.gz] from either PyKOfamSearch or PyHMMSearch (with header). Note: Assumes protein-level annotations have same protein identifiers as CDS sequences")
    parser.add_argument("-f","--annotation_format", type=str, default="pykofamsearch", 
                           choices={"pykofamsearch", "pyhmmsearch", "pykofamsearch-reformatted", "pyhmmsearch-reformatted", "veba-pfam","veba-kofam","veba-cazy","veba-uniref", "veba-mibig", "veba-vfdb","veba-amr", "custom"}, 
                           help = "Annotation format. If 'custom', then annotation table must be 2 columns [id_gene, id_feature] (No header, Tab delimited)"
                                   "and can contain multiple lines per gene. [Default: pykofamsearch]",
                                   )
    parser.add_argument("-o","--output_directory", type=str, help = "path/to/output_directory/ output directory that includes feature_mapping.tsv, genomes.tsv, and cds.fasta")
    # parser.add_argument("-f", "--force", action="store_true", help = "If --output_directory exists, then overwrite contents")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan preprocess")
    
    # Command
    logger.info(f"Command: {sys.argv}")

    # Output
    logger.info(f"Creating output directory: {opts.output_directory}")
    os.makedirs(opts.output_directory, exist_ok=True)

    # try:
        # os.makedirs(opts.output_directory, exist_ok=bool(opts.force))
    # except FileExistsError:
        # raise FileExistsError(f"--output_directory {opts.output_directory} already exists.  Either remove {opts.output_directory}, change --output_directory, or use --force to overwrite")

    # Input
    logger.info(f"Reading input: {opts.input}")
    if opts.input == "stdin":
        opts.input = sys.stdin

    df_input = pd.read_csv(opts.input, sep="\t", header=None)
    number_of_genomes, m = df_input.shape
    if m not in {3,4}:
        raise IndexError(f"--input has {m} columns but must be either 3 or 4 columns: [id_genome, path/to/assembly.fasta, path/to/cds.fasta. [Optional: id_genome_cluster]] (No header, Tab delimited)")
    df_input.columns = ["id_genome", "assembly", "cds", "id_genome_cluster"][:m]
    df_input = df_input.set_index("id_genome")
    
    genome_to_genomecluster = None
    if "id_genome_cluster" in df_input.columns:
        genome_to_genomecluster = df_input["id_genome_cluster"].to_dict()
        

    # Annotations
    logger.info(f"Reading annotations: {opts.annotations} [Format: {opts.annotation_format}]")
    gene_to_features = read_annotations(opts.annotations, format=opts.annotation_format)    
     
    # Build CDS fasta
    cds_identifiers = set()
    gene_to_genome = dict()
    with open_file_writer(os.path.join(opts.output_directory, "cds.fasta.gz")) as f:
        for id_genome, cds_filepath in tqdm(df_input["cds"].items(), total=number_of_genomes, desc="Extract feature sequences from CDS fasta"):
            for id_gene, seq in pyfastx.Fasta(cds_filepath, build_index=False):
                if id_gene in gene_to_features:
                    feature_set = gene_to_features[id_gene]
                    print(">{} [id_genome={}][features={}]\n{}".format(id_gene, id_genome, feature_set, seq), file=f)
                    cds_identifiers.add(id_gene)
                    gene_to_genome[id_gene] = id_genome
                    
    genes_missing_cds_sequences = set(gene_to_features.keys()) - cds_identifiers
    if genes_missing_cds_sequences:
        logger.warning(f"Number of genes in --annotations that do not have sequences from --input: {len(genes_missing_cds_sequences)}")
        with open_file_writer(os.path.join(opts.output_directory, "genes_missing_cds_sequences.list")) as f:
            for id_gene in genes_missing_cds_sequences:
                print(id_gene, file=f)
                
    if not cds_identifiers:
        raise IndexError("None of the CDS identifiers had protein representatives in --annotations")
    
    logger.info(f"Number of CDS sequences with features: {len(cds_identifiers)}")

    # Feature mapping
    feature_mapping_filepath = os.path.join(opts.output_directory, "feature_mapping.tsv.gz")
    logger.info(f"Building feature mappings: {feature_mapping_filepath}")

    f_feature_mapping = open_file_writer(feature_mapping_filepath)
    genomes_with_features = set()
    if genome_to_genomecluster:
        for id_gene, feature_set in tqdm(gene_to_features.items(), desc=f"Writing feature mapping table [Genome_Clusters = True]"): # [id_gene, feature_set, id_genome, (Optional: id_genome_cluster)] 
            if id_gene in cds_identifiers:
                id_genome = gene_to_genome[id_gene]
                id_genome_cluster = genome_to_genomecluster[id_genome]
                fields = [
                    id_gene, 
                    feature_set,
                    id_genome,
                    id_genome_cluster,
                ]
                print(*fields, sep="\t", file=f_feature_mapping)  
                genomes_with_features.add(id_genome)
    else:
        for id_gene, feature_set in tqdm(gene_to_features.items(), desc=f"Writing feature mapping table [Genome_Clusters = False]"): # [id_gene, feature_set, id_genome, (Optional: id_genome_cluster)] 
            if id_gene in cds_identifiers:
                id_genome = gene_to_genome[id_gene]
                fields = [
                    id_gene, 
                    feature_set,
                    id_genome,
                ]
                print(*fields, sep="\t", file=f_feature_mapping) 
                genomes_with_features.add(id_genome)

    f_feature_mapping.close()


    # Genomes
    genomes_with_filepaths = set(df_input.index)
    genomes_without_features = genomes_with_filepaths - genomes_with_features
    n_genomes_without_features = len(genomes_without_features)
    if genomes_without_features:
        logger.info(f"Including {df_input.shape[0] - n_genomes_without_features} genomes in index and excluding {n_genomes_without_features} genomes")
        genomes_excluded_filepath = os.path.join(opts.output_directory, "genomes_excluded.list")
        logger.info(f"Writing {n_genomes_without_features} excluded genomes: {genomes_excluded_filepath}")
        with open_file_writer(genomes_excluded_filepath) as f:
            for id in genomes_without_features:
                print(id, file=f)
    else:
        logger.info(f"Including all {df_input.shape[0]} genomes in index")

    genomes_filepath = os.path.join(opts.output_directory, "genomes.tsv.gz")
    logger.info(f"Writing genome filepaths: {genomes_filepath}")
    df_input = df_input.drop(list(genomes_without_features))
    df_input["assembly"].to_csv(genomes_filepath, sep="\t", header=None)

    # ========
    # Hash
    # ========   
    md5hash_filepath = os.path.join(opts.output_directory, "md5hashes.json")
    logger.info(f"Calculating md5 hashes: {md5hash_filepath}")
    file_to_hash = get_md5hash_from_directory(opts.output_directory)
    write_json(file_to_hash, md5hash_filepath)

    # ========
    # Complete
    # ========    
    logger.info(f"Completed preprocessing input data for leviathan index: {opts.output_directory}")
    logger.info(f"Directory size of leviathan preprocessing: {format_bytes(get_directory_size(opts.output_directory))}")
    logger.info(f"Directory structure of leviathan preprocessing:\n{get_directory_tree(opts.output_directory, ascii=True)}")
    logger.info(f"Completed running leviathan-preprocess: {opts.output_directory}")

if __name__ == "__main__":
    main()
    
    

    



#!/usr/bin/env python
import sys,os, argparse, warnings, subprocess, glob
from collections import defaultdict
import pandas as pd
import xarray as xr
# from pandas.errors import EmptyDataError
# from Bio.SeqIO.FastaIO import SimpleFastaParser
from tqdm import tqdm
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    # open_file_reader,
    # open_file_writer,
    read_pickle, 
    # write_pickle,
    read_json,
    # write_json,
    build_logger,
    # reset_logger,
    # format_duration,
    # format_header,
    get_file_size,
    format_bytes,
    # get_directory_tree,
    # get_directory_size,
    # get_md5hash_from_directory,
    RunShellCommand,
)

from leviathan.profile_taxonomy import(
    check_genome_database,
    check_reads_format,
    run_sylph_reads_sketcher,
    run_sylph_profiler,
)

def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} -1 forward.fq[.gz] -2 reverse.fq[.gz] -n sample_name -o project_directory --index_directory path/to/leviathan_index/"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser_io = parser.add_argument_group('I/O arguments')
    parser_io.add_argument("-1","--forward_reads", type=str,  help = "path/to/forward_reads.fq[.gz] (Cannot be used with -s/--read_sketch)")
    parser_io.add_argument("-2","--reverse_reads", type=str,  help = "path/to/reverse_reads.fq[.gz] (Cannot be used with -s/--read_sketch)")
    parser_io.add_argument("-s","--reads_sketch", type=str, help = "path/to/reads_sketch.sylsp (e.g., sylph sketch output) (Cannot be used with -1/--forward_reads and -2/--reverse_reads)")
    parser_io.add_argument("-n", "--name", type=str, required=True, help="Name of sample")
    parser_io.add_argument("-o","--project_directory", type=str, default="leviathan_output/profiling/taxonomy", help = "path/to/project_directory (e.g., leviathan_output/profiling/taxonomy]")
    parser_io.add_argument("-d","--index_directory", type=str, required=True, help = "path/to/index_directory/")
    parser_io.add_argument("-f","--output_format", type=str, choices={"tsv", "parquet"}, default="parquet", help = "Output format [Default: parquet]")

    # Utilities
    parser_utility = parser.add_argument_group('Utility arguments')
    parser_utility.add_argument("-p","--n_jobs", type=int, default=1,  help = "Number of threads to use.  Use -1 for all available. [Default: 1]")

    # Sylph
    parser_sylph_sketch = parser.add_argument_group('Sylph reads sketcher arguments (Fastq)')
    parser_sylph_sketch.add_argument("--sylph_executable", type=str, help="Sylph executable [Default: $PATH]")
    parser_sylph_sketch.add_argument("--sylph_k", type=int, choices={21,31}, default=31,  help="Sylph |  Value of k. Only k = 21, 31 are currently supported. [Default: 31]")
    parser_sylph_sketch.add_argument("--sylph_minimum_spacing", type=int,  default=30,  help="Sylph |  Minimum spacing between selected k-mers on the genomes [Default: 30]")
    parser_sylph_sketch.add_argument("--sylph_subsampling_rate", type=int, default=200,  help="Sylph | Subsampling rate.	[Default: 200]")
    parser_sylph_sketch.add_argument("--sylph_sketch_options", type=str, default="", help="Sylph | More options for `sylph sketch` (e.g. --arg=1 ) [Default: '']")

    parser_sylph_profile = parser.add_argument_group('Sylph profile arguments')
    parser_sylph_profile.add_argument("--sylph_profile_minimum_ani", type=float, default=95, help="Sylph profile | Minimum adjusted ANI to consider (0-100). [Default: 95]")
    parser_sylph_profile.add_argument("--sylph_profile_minimum_number_kmers", type=int, default=50, help="Sylph profile | Exclude genomes with less than this number of sampled k-mers.  Default is 50 in Sylph but lowering to 20 accounts for viruses and small CPR genomes. [Default: 50]")
    parser_sylph_profile.add_argument("--sylph_profile_minimum_count_correct", type=int, default=3, help="Sylph profile | Minimum k-mer multiplicity needed for coverage correction. Higher values gives more precision but lower sensitivity [Default: 3]")
    parser_sylph_profile.add_argument("--sylph_profile_options", type=str, default="", help="Sylph profile | More options for `sylph profile` (e.g. --arg 1 ) [Default: '']")


    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan profile-taxonomy")

    # Commands
    logger.info(f"Command: {sys.argv}")

    # Threads
    if opts.n_jobs == -1:
        from multiprocessing import cpu_count 
        opts.n_jobs = cpu_count()
        logger.info(f"Setting --n_jobs to maximum threads {opts.n_jobs}")

    assert opts.n_jobs >= 1, "--n_jobs must be ??? 1.  To select all available threads, use -1."
    
    # Executables
    # * Sylph
    if not opts.sylph_executable:
        opts.sylph_executable = os.path.join(bin_directory, "sylph")
    if not os.path.exists(opts.sylph_executable):
        msg = f"sylph executable not doesn't exist: {opts.sylph_executable}"
        logger.critical(msg)
        raise FileNotFoundError(msg)
    
    # Config
    config = read_json(os.path.join(opts.index_directory, "config.json"))

    # Determine input reads format
    input_reads_format = check_reads_format(
        forward_reads=opts.forward_reads, 
        reverse_reads=opts.reverse_reads, 
        reads_sketch=opts.reads_sketch,
        logger=logger,
        )

    # Check Sylph database
    check_genome_database(
        index_directory=opts.index_directory, 
        logger=logger,
        )
    
    # Output
    output_directory = os.path.join(opts.project_directory, opts.name)
    os.makedirs(output_directory, exist_ok=True)
    os.makedirs(os.path.join(output_directory, "output"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "intermediate"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "logs"), exist_ok=True)
    os.makedirs(os.path.join(output_directory, "tmp"), exist_ok=True)

    if input_reads_format == "paired":
        # Process and check inputs
        logger.info("Sketching reads")

        # ==========================
        # Build Sylph reads sketcher
        # ==========================
        cmd_sylph_reads_sketcher = run_sylph_reads_sketcher(
                        logger=logger,
                        log_directory=os.path.join(output_directory, "logs"), 
                        sylph_executable=opts.sylph_executable, 
                        n_jobs=opts.n_jobs, 
                        output_directory=os.path.join(output_directory, "intermediate"), 
                        forward_reads=opts.forward_reads, 
                        reverse_reads=opts.reverse_reads,
                        k=opts.sylph_k, 
                        minimum_spacing=opts.sylph_minimum_spacing, 
                        subsampling_rate=opts.sylph_subsampling_rate, 
                        sylph_sketch_options=opts.sylph_sketch_options, 

        )
        opts.reads_sketch = os.path.join(output_directory, "intermediate", "reads.sylsp")
        logger.info(f"Setting --reads_sketch to {opts.reads_sketch}")
    else:
        logger.info(f"Skipping read sketching and using {opts.reads_sketch}")
        if opts.reads_sketch is None:
            raise ValueError("Please either provide -1/-2 paired reads or -s/--reads_sketch")

        
    # ====================
    # Build Sylph profiler
    # ====================
    # Run Sylph
    logger.info("Running Sylph profiler")
    cmd_sylph_profiler = run_sylph_profiler(
        logger=logger,
        log_directory=os.path.join(output_directory, "logs"), 
        sylph_executable=opts.sylph_executable, 
        n_jobs=opts.n_jobs, 
        output_directory=os.path.join(output_directory, "output"), 
        index_directory=opts.index_directory,
        reads=opts.reads_sketch, 
        minimum_ani=opts.sylph_profile_minimum_ani, 
        minimum_number_kmers=opts.sylph_profile_minimum_number_kmers, 
        minimum_count_correct=opts.sylph_profile_minimum_count_correct, 
        sylph_profile_options=opts.sylph_profile_options, 
    )

    # =======================
    # Process genome metadata
    # =======================
    logger.info("Processing genome metadata")
    
    genomefilepath_to_genomeid = dict()
    genome_to_genomecluster = dict()
    
    genome_to_data = read_pickle(os.path.join(opts.index_directory, "database", "genome_to_data.pkl.gz"))
    for id_genome, data in tqdm(genome_to_data.items(), "Getting filepaths associated with genomes"):
        filepath = data["filepath"] 
        genomefilepath_to_genomeid[filepath] = id_genome
        genome_to_genomecluster[id_genome] = data["id_genome_cluster"]
    del genome_to_data
       
    # ===================================
    # Reformat taxonomic abundance tables
    # ===================================
    genome_abundance_filepath = os.path.join(output_directory, "output", f"taxonomic_abundance.genomes.{opts.output_format}")
    if opts.output_format != "parquet":
        genome_abundance_filepath += ".gz"

    
    logger.info(f"Reformatting taxonomic abundance: {genome_abundance_filepath}")
    df_sylph = pd.read_csv(os.path.join(output_directory, "output", "sylph_profile.tsv.gz"), sep="\t", index_col=0)
    genome_to_abundance = df_sylph.reset_index().set_index("Genome_file")["Taxonomic_abundance"]
    genome_to_abundance.index = genome_to_abundance.index.map(lambda filepath:genomefilepath_to_genomeid[filepath])
    genome_to_abundance.index.name = "id_genome"
    genome_to_abundance = genome_to_abundance.to_frame("abundance")
    if opts.output_format == "parquet":
        genome_to_abundance.to_parquet(genome_abundance_filepath, index=True)
    elif opts.output_format == "tsv":
        genome_to_abundance.to_csv(genome_abundance_filepath, sep="\t")
    
    if config["contains_genome_cluster_mapping"]:
        genomecluster_abundance_filepath = os.path.join(output_directory, "output", f"taxonomic_abundance.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            genomecluster_abundance_filepath += ".gz"
        logger.info(f"Aggregating taxonomic abundance for genome clusters: {genomecluster_abundance_filepath}")
        genomecluster_to_abundance = genome_to_abundance["abundance"].groupby(genome_to_genomecluster).sum()
        genomecluster_to_abundance.index.name = "id_genome_cluster"
        genomecluster_to_abundance = genomecluster_to_abundance.to_frame("abundance")
        if opts.output_format == "parquet":
            genomecluster_to_abundance.to_parquet(genomecluster_abundance_filepath, index=True)
        elif opts.output_format == "tsv":
            genomecluster_to_abundance.to_csv(genomecluster_abundance_filepath, sep="\t")

    # ===================================
    # Reformat sequence abundance tables
    # ===================================
    genome_abundance_filepath = os.path.join(output_directory, "output", f"sequence_abundance.genomes.{opts.output_format}")
    if opts.output_format != "parquet":
        genome_abundance_filepath += ".gz"

    
    logger.info(f"Reformatting sequence abundance: {genome_abundance_filepath}")
    df_sylph = pd.read_csv(os.path.join(output_directory, "output", "sylph_profile.tsv.gz"), sep="\t", index_col=0)
    genome_to_abundance = df_sylph.reset_index().set_index("Genome_file")["Sequence_abundance"]
    genome_to_abundance.index = genome_to_abundance.index.map(lambda filepath:genomefilepath_to_genomeid[filepath])
    genome_to_abundance.index.name = "id_genome"
    genome_to_abundance = genome_to_abundance.to_frame("abundance")
    if opts.output_format == "parquet":
        genome_to_abundance.to_parquet(genome_abundance_filepath, index=True)
    elif opts.output_format == "tsv":
        genome_to_abundance.to_csv(genome_abundance_filepath, sep="\t")
    
    if config["contains_genome_cluster_mapping"]:
        genomecluster_abundance_filepath = os.path.join(output_directory, "output", f"sequence_abundance.genome_clusters.{opts.output_format}")
        if opts.output_format != "parquet":
            genomecluster_abundance_filepath += ".gz"
        logger.info(f"Aggregating sequence abundance for genome clusters: {genomecluster_abundance_filepath}")
        genomecluster_to_abundance = genome_to_abundance["abundance"].groupby(genome_to_genomecluster).sum()
        genomecluster_to_abundance.index.name = "id_genome_cluster"
        genomecluster_to_abundance = genomecluster_to_abundance.to_frame("abundance")
        if opts.output_format == "parquet":
            genomecluster_to_abundance.to_parquet(genomecluster_abundance_filepath, index=True)
        elif opts.output_format == "tsv":
            genomecluster_to_abundance.to_csv(genomecluster_abundance_filepath, sep="\t")
            
    # ========
    # Complete
    # ========    
    logger.info(f"Completed taxonomic profiling: {opts.name}")
    for filepath in glob.glob(os.path.join(output_directory, "output","*")):
        filesize = get_file_size(filepath, format=True)
        logger.info(f"Output: {filepath} ({filesize})")
    logger.info(f"Completed running leviathan-profile-pathway for {opts.name}: {opts.project_directory}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python
import sys
import os
import glob
import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm
from collections import OrderedDict
# from memory_profiler import profile

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    format_bytes,
    get_file_size,
    RunShellCommand,
    check_argument_choice,
)

    
def check_reads_format(forward_reads, reverse_reads, reads_sketch, logger):
    input_reads_format = None
    if any([forward_reads, reverse_reads]):
        assert forward_reads != reverse_reads, f"You probably mislabeled the input files because `forward_reads` should not be the same as `reverse_reads`: {forward_reads}"
        assert forward_reads is not None, "If running in --input_reads_format paired mode, --forward_reads and --reverse_reads are needed."
        assert reverse_reads is not None, "If running in --input_reads_format paired mode, --forward_reads and --reverse_reads are needed."
        input_reads_format = "paired"
    if reads_sketch is not None:
        assert forward_reads is None, "If running in --input_reads_format sketch mode, you cannot provide --forward_reads, --reverse_reads"
        assert reverse_reads is None, "If running in --input_reads_format sketch mode, you cannot provide --forward_reads, --reverse_reads"
        input_reads_format = "sketch"
    if input_reads_format is None:
        msg = "Could not determine input reads format.  Please provide either paired fastq or a Sylph sketch."
        logger.critical(msg)
        raise ValueError(msg)
    logger.info(f"Auto-detecting reads format: {input_reads_format}")
    return input_reads_format

def check_genome_database(index_directory, logger):
    # Sylph
    genomes_database_filepath = os.path.join(index_directory, "database", "genomes.syldb")
    if not os.path.exists(genomes_database_filepath):
        msg = f"Could not find the following genomes database: {genomes_database_filepath}. Please ensure you used --genomes when building index with `leviathan index`.  If not, you can update using the --update_with_genomes argument and retry."
        logger.critical(msg)
        raise FileNotFoundError(msg)
    else:
        filesize = get_file_size(genomes_database_filepath, format=True)
        logger.info(f"The following genomes database was found: {genomes_database_filepath} ({filesize})")
        
    # Pickle
    genomes_data_filepath = os.path.join(index_directory, "database", "genome_to_data.pkl.gz")
    if not os.path.exists(genomes_data_filepath):
        msg = f"Could not find the following genomic data: {genomes_data_filepath}. Please ensure you used --genomes when building index with `leviathan index`.  If not, you can update using the --update_with_genomes argument and retry."
        logger.critical(msg)
        raise FileNotFoundError(msg)
    else:
        filesize = get_file_size(genomes_data_filepath, format=True)
        logger.info(f"The following genomes database was found: {genomes_data_filepath} ({filesize})")
    
        
# Run Sylph reads sketcher
def run_sylph_reads_sketcher(logger, log_directory, sylph_executable, n_jobs, output_directory, forward_reads, reverse_reads, k, minimum_spacing, subsampling_rate, sylph_sketch_options):
    forward_reads_filename = os.path.split(forward_reads)[-1]
    cmd = RunShellCommand(
        command=[
            sylph_executable,
            "sketch",
            "-t",
            n_jobs,
            "-k",
            k,
            "-c",
            subsampling_rate,
            "--min-spacing",
            minimum_spacing,
            "-d",
            output_directory,
            "-1",
            forward_reads,
            "-2",
            reverse_reads,
            "&&",
            "mv",
            os.path.join(output_directory, f"{forward_reads_filename}.paired.sylsp"),
            os.path.join(output_directory, "reads.sylsp"),
            
        ],
        name="sylph_reads_sketcher",
        validate_input_filepaths=[
            forward_reads,
            reverse_reads,
        ],
        validate_output_filepaths=[
            forward_reads,
            reverse_reads,
            os.path.join(output_directory, "reads.sylsp"),
        ],
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd

# Run Sylph profile
def run_sylph_profiler(logger, log_directory, sylph_executable, n_jobs, output_directory, index_directory,  reads, minimum_ani, minimum_number_kmers, minimum_count_correct, sylph_profile_options):
    cmd = RunShellCommand(
        command=[
            sylph_executable,
            "profile",
            "--estimate-unknown",
            "-t",
            n_jobs,
            "--minimum-ani",
            minimum_ani,
            "--min-number-kmers",
            minimum_number_kmers,
            "--min-count-correct",
            minimum_count_correct,
            sylph_profile_options,
            os.path.join(index_directory, "database", "genomes.syldb"),
            reads,
            "|",
            "gzip", # pigz?
            ">",
            os.path.join(output_directory, "sylph_profile.tsv.gz"),
        ],
        name="sylph_profiler",
        validate_input_filepaths=[
            os.path.join(index_directory, "database", "genomes.syldb"),
            reads,
        ],
        validate_output_filepaths=[
            os.path.join(output_directory, "sylph_profile.tsv.gz"),
        ]
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd

def merge_taxonomic_profiling_tables_as_pandas(profiling_directory:str, level="genome", data_type:str="taxonomic_abundances", fillna_with_zeros:bool=False, sparse:bool=False):
    
    """
    Merge taxonomic profiling at the genome or genome cluster level.

    Parameters
    ----------
    profiling_directory : str
        Directory containing the profiling output.
    level : str, optional
        Level at which to merge the taxonomic profiling. Options are
        "genome" or "genome_cluster". Default is "genome".
    data_type : str, optional
        Abundance type either `taxonomic_abundances` or `sequence_abundances` Default is "taxonomic_abundances".
    fillna_with_zeros : bool, optional
        Whether to fill missing values with zeros. Default is False.
    sparse : bool, optional
        Whether to return a pd.Sparse type. Default is False.
        
    Returns
    -------
    pd.DataFrame
        Merged taxonomic profiling at the specified level.

    Raises
    ------
    KeyError
        If `level` is not in {"genome", "genome_cluster"}.
    FileNotFoundError
        If no files are found in the specified directory.

    Files:
    * taxonomic_abundance.genome_clusters.parquet
    * taxonomic_abundance.genomes.parquet
    """
    choices = {"genomes", "genome_clusters"}
    if level not in choices:
        raise KeyError(f"level must be in {choices}")
    

    # Genomes
    if level == "genomes":
        output = dict()

        filepaths = glob.glob(f"{profiling_directory}/*/output/{data_type[:-1]}.genomes.parquet")

        if filepaths:
            for filepath in tqdm(filepaths, f"Merging {level}-level {data_type.replace("_", " ")}"):
                id_sample = filepath.split("/")[-3]
                output[id_sample] = pd.read_parquet(filepath).squeeze("columns")
        else:
            raise FileNotFoundError(f"Could not find any {data_type[:-1]}.genomes.parquet files in {profiling_directory}")
     
    # Genome clusters
    elif level == "genome_clusters":
        output = dict()

        filepaths = glob.glob(f"{profiling_directory}/*/output/{data_type[:-1]}.genome_clusters.parquet")

        if filepaths:
            for filepath in tqdm(filepaths, f"Merging {level}-level {data_type.replace("_", " ")}"):
                id_sample = filepath.split("/")[-3]
                # output[id_sample] = pd.read_csv(filepath, sep="\t", index_col=0).squeeze("columns")
                output[id_sample] = pd.read_parquet(filepath).squeeze("columns")
        else:
            raise FileNotFoundError(f"Could not find any {data_type[:-1]}.genome_clusters.parquet files in {profiling_directory}")
        
    X = pd.DataFrame(output).T
    missing_value_fill=pd.NA
    if fillna_with_zeros:
        X = X.fillna(0.0)
        missing_value_fill = 0.0
    if sparse:
        X = X.astype(pd.SparseDtype("float", missing_value_fill))
    return X

def merge_taxonomic_profiling_tables_as_xarray(profiling_directory:str, level="genomes", fillna_with_zeros:bool=False):
    
    """
    merges sample-level {data_type} values from multiple samples into a single DataFrame.

    Parameters
    ----------
    profiling_directory : str
        Path to directory containing sample-level directories with output files.
    level : str, optional
        Level of organization for {data_type}. One of {"genomes", "genome_cluster"}.
    fillna_with_zeros : bool, optional
        Whether to fill missing values with zeros. Default is False.

    Returns
    -------
    xr.DataArray
        Merged xr.DataArray with dims: (samples, {level}, features)

    Notes
    -----
    
    Files:
    * taxonomic_abundances.genomes.parquet
    * taxonomic_abundances.genome_clusters.parquet

    """


    check_argument_choice(
        query=level, 
        choices={"genomes", "genome_clusters"},
        )

    # Output data
    taxonomic_abundance_filepaths = glob.glob(f"{profiling_directory}/*/output/taxonomic_abundance.{level}.parquet")
    sequence_abundance_filepaths = glob.glob(f"{profiling_directory}/*/output/sequence_abundance.{level}.parquet")

    if not taxonomic_abundance_filepaths:
        raise FileNotFoundError(f"Could not find any taxonomic_abundance.{level}.parquet files in {profiling_directory}")
    if not sequence_abundance_filepaths:
        raise FileNotFoundError(f"Could not find any sequence_abundance.{level}.parquet files in {profiling_directory}")


    # Merge abundances
    output = OrderedDict()
    for variable_label in ["taxonomic_abundances", "sequence_abundances"]:
        df = merge_taxonomic_profiling_tables_as_pandas(
            profiling_directory=profiling_directory, 
            level=level, 
            data_type=variable_label, 
            fillna_with_zeros=fillna_with_zeros, 
            sparse=False,
        )
        output[variable_label] = xr.DataArray(data = df.values, coords = [("samples", df.index), (level, df.columns)])
        del df
                
      
    X = xr.Dataset(output)
    X = X.astype(np.float32)
    if fillna_with_zeros:
        X = X.fillna(0.0)
    return X

    
    

    



from setuptools import setup
import os

script_directory = os.path.abspath(os.path.dirname(__file__))

package_name = "leviathan"
version = None
with open(os.path.join(script_directory, package_name, '__init__.py')) as f:
    for line in f.readlines():
        line = line.strip()
        if line.startswith("__version__"):
            version = line.split("=")[-1].strip().strip('"')
assert version is not None, f"Check version in {package_name}/__init__.py"

requirements = list()
with open(os.path.join(script_directory, 'requirements.txt')) as f:
    for line in f.readlines():
        line = line.strip()
        if line:
            if not line.startswith("#"):
                requirements.append(line)
                
setup(name='leviathan',
    version=version,
    description='Genome-resolved taxonomic and pathway profiling',
    url='https://github.com/jolespin/leviathan',
    author='Josh L. Espinoza',
    author_email='jol.espinoz@gmail.com',
    license='Apache 2.0',
    packages=["leviathan"],
    install_requires=requirements,
    include_package_data=False,
    scripts=[
        "bin/compile-manifest-from-veba.py",
        "bin/leviathan-preprocess.py",
        "bin/leviathan-index.py",
        "bin/leviathan-info.py",
        "bin/leviathan-merge.py",
        "bin/leviathan-profile-taxonomy.py",
        "bin/leviathan-profile-pathway.py",
    ],

)




#!/usr/bin/env python
import sys, os, argparse, glob
from collections import defaultdict
from tqdm import tqdm

__program__ = os.path.split(sys.argv[0])[-1]

from pyexeggutor import (
    open_file_reader,
    open_file_writer,
    build_logger,
)


def main(args=None):
    # Options
    # =======
    # Path info
    python_executable = sys.executable
    bin_directory = "/".join(python_executable.split("/")[:-1])
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, sys.version.split(" ")[0], python_executable, script_filename)
    usage = f"{__program__} -i path/to/veba_output/ -t prokaryotic,eukaryotic -o path/to/output.tsv"
    epilog = "Leviathan"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser_io = parser.add_argument_group('I/O arguments')
    parser_io.add_argument("-i","--veba_directory", type=str, required=True, help = "path/to/veba_directory/ (e.g., veba_output/)")
    parser_io.add_argument("-t","--organism_types", type=str, default = "prokaryotic,eukaryotic", help="Comma-separated list of organism types.  Choose between {prokaryotic, eukaryotic, viral}(e.g., prokaryotic,eukaryotic).  viral is not recommended.")
    parser_io.add_argument("-o", "--output", type=str,  default="stdout", help = "path/to/output.tsv[.gz] [Default: stdout]")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # logger
    logger = build_logger("leviathan compile-manifest-from-veba")

    # Commands
    logger.info(f"Command: {sys.argv}")
     
    # Inputs
    logger.info(f"Checking input directory: {opts.veba_directory}")

    if not os.path.exists(opts.veba_directory):
        logger.error(f"Error: {opts.veba_directory} does not exist.")
        sys.exit(1)
    if not os.path.isdir(opts.veba_directory):
        logger.error(f"Error: {opts.veba_directory} is not a directory.")
        sys.exit(1)
    # Output
    if opts.output == "stdout":
        f_out = sys.stdout
    else:
        f_out = open_file_writer(opts.output)
        
    logger.info(f"Creating output: {f_out}")
    
    # Organism types
    opts.organism_types = opts.organism_types.split(",")
    logger.info(f"Parsing organism types: {opts.organism_types}")

    for organism_type in opts.organism_types:
        if organism_type.lower() == "viral":
            logger.warning(f"'viral' is not recommended for organism types because small genomes require different parameters than medium-to-large genomes with Sylph.")

    # Create manifest
    manifest = defaultdict(dict)
    for organism_type in opts.organism_types:
        organism_type_directory = os.path.join(opts.veba_directory, "binning", organism_type)
        logger.info(f"Creating manifest from {organism_type_directory}")
        if not os.path.exists(organism_type_directory):
            logger.critical(f"{organism_type_directory} does not exist.")
            sys.exit(1)
        if not os.path.isdir(organism_type_directory):
            logger.error(f"{organism_type_directory}  is not a directory.")
            sys.exit(1)
        for filepath in tqdm(glob.glob(os.path.join(organism_type_directory, "*", "output", "genomes", "*.fa")), f"Identifying genomes assemblies for {organism_type}", unit=" genomes"):
            id_genome = os.path.split(filepath)[1][:-3]
            manifest[id_genome]["assembly"] = filepath
        for filepath in tqdm(glob.glob(os.path.join(organism_type_directory, "*", "output", "genomes", "*.ffn")), f"Identifying CDS sequences for {organism_type}", unit=" genomes"):
            id_genome = os.path.split(filepath)[1][:-4]
            manifest[id_genome]["cds"] = filepath
        
    # Genome clusters
    contains_genome_clusters = False
    genome_cluster_filepath = os.path.join(opts.veba_directory, "cluster", "output", "global", "mags_to_slcs.tsv")
    
    if os.path.exists(genome_cluster_filepath):
        logger.info(f"Adding genome clusters from {genome_cluster_filepath}")
        with open_file_reader(genome_cluster_filepath) as f_in:
            for line in f_in:
                line = line.strip()
                if line:
                    id_genome, id_cluster = line.strip().split("\t")
                    if id_genome in manifest:
                        manifest[id_genome]["id_genome_cluster"] = id_cluster
        contains_genome_clusters = True
    else:
        logger.warning(f"No genome clusters found at {genome_cluster_filepath}")

    

    # Check files
    task_description = f"Checking assembly and CDS files for manifest"
    logger.info(task_description)
    for id_genome, data in tqdm(manifest.items(), task_description, unit=" genomes"):
        try:
            filepath = data["assembly"]
            if not os.path.exists(filepath):
                logger.critical(f"{filepath} does not exist.")
                sys.exit(1)
        except KeyError:
            logger.critical(f"{id_genome} does not have an assembly fasta file.")
            sys.exit(1)
        try:
            filepath = data["cds"]
            if not os.path.exists(filepath):
                logger.critical(f"{filepath} does not exist.")
                sys.exit(1)
        except KeyError:
            logger.critical(f"{id_genome} does not have a CDS fasta file.")
            sys.exit(1)
    if contains_genome_clusters:
        task_description = f"Checking clustering for genomes in manifest"
        logger.info(task_description)
        genomes_missing_clusters = set()
        for id_genome, data in tqdm(manifest.items(), task_description, unit=" genomes"):
            try:
                id_cluster = data["id_genome_cluster"]
            except KeyError:
                genomes_missing_clusters.add(id_genome)
        if genomes_missing_clusters:
            logger.critical("The following genomes do not have genome clusters:\n{}".format("\n".join(sorted(genomes_missing_clusters))))
            sys.exit(1)
            
    # Write manifest
    logger.info(f"Writing manifest: {f_out}")
    if contains_genome_clusters:
        for id_genome, data in manifest.items():
            print(id_genome, data["assembly"], data["cds"], data["id_genome_cluster"],sep="\t", file=f_out)
    else:
        for id_genome, data in manifest.items():
            print(id_genome, data["assembly"], data["cds"], sep="\t", file=f_out)
    if f_out != sys.stdout:
        f_out.close()
        
    # ========
    # Complete
    # ========    
    logger.info(f"Completed generating manifest: {f_out}")
            
if __name__ == "__main__":
    main()
    
    

    



#!/usr/bin/env python
import sys
import os
import glob
from collections import defaultdict
from tqdm import tqdm
import pandas as pd
import numpy as np
import xarray as xr
# from memory_profiler import profile

from kegg_pathway_profiler.pathways import (
    pathway_coverage_wrapper,
)

from pyexeggutor import (
    format_bytes,
    RunShellCommand,
    check_argument_choice,
)
    
        
        
__program__ = os.path.split(sys.argv[0])[-1]


# Run Salmon quant (salmon quant --meta --libType A --index ${INDEX} -1 ${R1} -2 ${R2} --threads ${N_JOBS}  --minScoreFraction=0.87 --writeUnmappedNames)
def run_salmon_quant(logger, log_directory, salmon_executable, samtools_executable, n_jobs, output_directory, index_directory, forward_reads, reverse_reads, minimum_score_fraction, include_mappings, alignment_format, salmon_gzip, salmon_quant_options):
    arguments = dict(
        command=[
            salmon_executable,
            "quant",
            "--meta",
            "--libType",
            "A",
            "--threads",
            n_jobs,
            "--minScoreFraction",
            minimum_score_fraction,
            "--index",
            os.path.join(index_directory, "salmon_index"),
            "-1",
            forward_reads,
            "-2",
            reverse_reads,
            "--writeUnmappedNames",
            salmon_quant_options if salmon_quant_options else "", 
            "--output",
            output_directory,
        ],
        name="salmon_quant",
        validate_input_filepaths=[
            forward_reads,
            reverse_reads,
        ],
        validate_output_filepaths=[
            os.path.join(output_directory, "quant.sf.gz") if salmon_gzip else os.path.join(output_directory, "quant.sf"),
        ],
    )
    if include_mappings:
        if alignment_format == "sam":
            arguments["command"] += [
                "--writeMappings",
                ">",
                os.path.join(output_directory, "mapped.sam"),
            ]
            arguments["validate_output_filepaths"].append(os.path.join(output_directory, "mapped.sam"))
            
        elif alignment_format == "bam":
            arguments["command"] += [
                "--writeMappings",
                "|",
                samtools_executable,
                "view",
                "-b",
                "-h",
                "-o",
                os.path.join(output_directory, "mapped.bam"),
            ]
            arguments["validate_output_filepaths"].append(os.path.join(output_directory, "mapped.bam"))

        elif alignment_format == "sorted.bam":
            arguments["command"] += [
                "--writeMappings",
                "|",
                samtools_executable,
                "view",
                "-b",
                "-h",
                "|",
                samtools_executable,
                "sort",
                "-@",
                n_jobs,
                "-o",
                os.path.join(output_directory, "mapped.sorted.bam"),
                "-",
            ]
            arguments["validate_output_filepaths"].append(os.path.join(output_directory, "mapped.sorted.bam"))
            
    # Remove unmapped reads
    arguments["command"] += [
        "&&",
        "rm",
        "-v",
        os.path.join(output_directory, "aux_info", "unmapped_names.txt"),
    ]
        
    # Gzip quant.sf
    if salmon_gzip:
        arguments["command"] += [
            "&&",
            "gzip",
            os.path.join(output_directory, "quant.sf"),
        ]
    
    cmd = RunShellCommand(
        **arguments,
    )
    
    # Run
    logger.info(f"[{cmd.name}] running command: {cmd.command}")
    cmd.run()
    logger.info(f"[{cmd.name}] duration: {cmd.duration_}")
    logger.info(f"[{cmd.name}] peak memory: {format_bytes(cmd.peak_memory_)}")

    # Dump
    logger.info(f"[{cmd.name}] dumping stdout, stderr, and return code: {log_directory}")
    cmd.dump(log_directory)
    
    # Validate
    logger.info(f"[{cmd.name}] checking return code status: {cmd.returncode_}")
    cmd.check_status()
    return cmd

def reformat_gene_abundance(df_quant:pd.DataFrame, gene_to_data:dict):
    # Name    Length  EffectiveLength TPM     NumReads
    index = list()
    values = list()
    for id_gene, row in tqdm(df_quant.iterrows(), "Removing zero-abundance features"):
        abundance = row["NumReads"]
        if abundance > 0:
            id_genome = gene_to_data[id_gene]["id_genome"]
            tpm = row["TPM"]
            values.append([abundance, tpm])
            index.append((id_genome, id_gene))
    return pd.DataFrame(
        data=values,
        index=pd.MultiIndex.from_tuples(index, names=["id_genome", "id_gene"]),
        columns=[ "number_of_reads", "tpm"],
    )
    
def reformat_feature_abundance(df_gene_abundances:pd.DataFrame, gene_to_data:dict, split_feature_abundances:bool):
    feature_to_values = defaultdict(lambda: np.zeros(2, dtype=float))
    if split_feature_abundances:
        columns = ["number_of_reads(scaled)", "tpm(scaled)"]
        for (id_genome, id_gene), row in tqdm(df_gene_abundances.iterrows(), "Aggregating feature counts (Splitting abundances across features)"):
            abundance, tpm = row
            features = gene_to_data[id_gene]["features"]
            if features:
                number_of_features = len(features)
                tpm_scaled = tpm/number_of_features
                abundance_scaled = abundance/number_of_features
                for id_feature in features:
                    feature_to_values[(id_genome, id_feature)] += [abundance_scaled, tpm_scaled]
    else:
        columns = ["number_of_reads", "tpm"]
        for (id_genome, id_gene), row in tqdm(df_gene_abundances.iterrows(), "Aggregating feature counts"):
            tpm, abundance = row
            features = gene_to_data[id_gene]["features"]
            if features:
                number_of_features = len(features)
                for id_feature in features:
                    feature_to_values[(id_genome, id_feature)] += [abundance_scaled, tpm_scaled]

                    
    df_output = pd.DataFrame(
        data=feature_to_values,
        index=columns,
    ).T
    df_output.index.names = ["id_genome", "id_feature"]
    return df_output
    
def build_wide_feature_prevalence_matrix(df_feature_abundance:pd.DataFrame, threshold:float=0.0):
    # Index
    genomes = sorted(df_feature_abundance.index.get_level_values(0).unique())
    features = sorted(df_feature_abundance.index.get_level_values(1).unique())
    # Shape
    number_of_genomes = len(genomes)
    number_of_features = len(features)
    # Lookup
    genome_index_lookup = dict(zip(genomes, range(number_of_genomes)))
    feature_index_lookup = dict(zip(features, range(number_of_features)))
    # Zero-matrix
    prevalence_matrix = np.zeros((number_of_genomes, number_of_features), dtype=int)
    # Populate zero-matrix
    for (id_genome, id_feature), value in tqdm(df_feature_abundance.iloc[:,0].items(), total=df_feature_abundance.shape[0]):
        if value > threshold:
            i = genome_index_lookup[id_genome]
            j = feature_index_lookup[id_feature]
            prevalence_matrix[i,j] += 1
      
    return pd.DataFrame(
        data=prevalence_matrix,
        index=pd.Index(genomes, name="id_genome"),
        columns=pd.Index(features, name="id_feature"),
    )
                    
def build_feature_prevalence_dictionary(df_feature_prevalence_binary:pd.DataFrame):
    genome_to_features = dict()
    for id_genome, prevalence in df_feature_prevalence_binary.iterrows():
        genome_to_features[id_genome] = set(prevalence[lambda x: x > 0].index)
    return genome_to_features

def build_feature_pathway_dictionary(pathway_to_data:dict):
    feature_to_pathways = defaultdict(set)
    for id_pathway, data in pathway_to_data.items():
        for id_feature in data["ko_to_nodes"]:
            feature_to_pathways[id_feature].add(id_pathway)
    return feature_to_pathways
            
        
def calculate_pathway_coverage(genome_to_features:dict, pathway_to_data:dict):
    # Coverage
    coverages = dict()
    # Calculate pathway coverage for all genomes
    for id_genome, evaluation_features in genome_to_features.items():
        # Calculate pathway coverage for all pathways based on evaluation feature set
        pathway_to_results = pathway_coverage_wrapper(
            evaluation_kos=evaluation_features,
            database=pathway_to_data,
            progressbar_description=f"Calculating pathway coverage: {id_genome}",
        )

        # Coverage
        for id_pathway, results in pathway_to_results.items():
            coverages[(id_genome, id_pathway)] = results["coverage"]
    
    return coverages

def aggregate_pathway_abundance_and_append_coverage(df_feature_abundance:pd.DataFrame, feature_to_pathways:dict, coverages:dict, index_names = ["id_genome", "id_pathway"]):
    abundance_matrix = defaultdict(lambda: np.zeros(3, dtype=float))
    for (id_genome, id_feature), values in df_feature_abundance.iterrows():
        for id_pathway in feature_to_pathways[id_feature]:
            abundance_matrix[(id_genome, id_pathway)][:-1] += values
    for (id_genome, id_pathway) in abundance_matrix:
        coverage = coverages.get((id_genome, id_pathway), 0.0)
        abundance_matrix[(id_genome, id_pathway)] += [0.0, 0.0, coverage]
            
    df_output = pd.DataFrame(abundance_matrix, index=df_feature_abundance.columns.tolist() + ["coverage"]).T
    df_output.index.names = index_names
    return df_output

def aggregate_feature_abundance_for_clusters(df_feature_abundance:pd.DataFrame, genome_to_data:dict):
    def f(x):
        id_genome, id_feature = x
        return (genome_to_data[id_genome]["id_genome_cluster"], id_feature)
    df_output = df_feature_abundance.groupby(f).sum()
    df_output.index = pd.MultiIndex.from_tuples(df_output.index, names=["id_genome_cluster", "id_feature"])
    return df_output

def merge_pathway_profiling_tables_as_pandas(profiling_directory:str, data_type:str, level="genomes", metric="number_of_reads", fillna_with_zeros:bool=False, sparse:bool=False):
    
    """
    merges sample-level {data_type} values from multiple samples into a single DataFrame.

    Parameters
    ----------
    profiling_directory : str
        Path to directory containing sample-level directories with output files.
    data_type : str
        Type of {level}-level data to merge. One of: {"feature_abundances", "feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio", "gene_abundances", "pathway_abundances"}
    level : str, optional
        Level of organization for {data_type}. One of {"genomes", "genome_cluster"}.
    metric : str, optional
        Metric to use for {data_type}. One of {"number_of_reads", "tpm", "coverage"}.
    fillna_with_zeros : bool, optional
        Whether to fill missing values with zeros. Default is False.
    sparse : bool, optional
        Whether to return a pd.Sparse type. Default is False. 
    Returns
    -------
    pd.DataFrame
        merged DataFrame with {data_type} values for each sample.

    Notes
    -----
    Will raise a ValueError if an invalid combination of arguments is provided, such as level="genome_cluster" and data_type="gene_abundances".
    
    Files:
    * feature_abundances.genome_clusters.parquet
    * feature_abundances.genomes.parquet
    * feature_prevalence-binary.genome_clusters.parquet
    * feature_prevalence-binary.genomes.parquet
    * feature_prevalence.genome_clusters.parquet
    * feature_prevalence.genomes.parquet
    * feature_prevalence-ratio.genome_clusters.parquet
    * gene_abundances.genomes.parquet
    * pathway_abundances.genome_clusters.parquet
    * pathway_abundances.genomes.parquet
    """

    check_argument_choice(
        query=data_type, 
        choices={"feature_abundances", "feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio", "gene_abundances", "pathway_abundances"},
        )
    check_argument_choice(
        query=level, 
        choices={"genomes", "genome_clusters"},
        )
    check_argument_choice(
        query=metric, 
        choices={"number_of_reads", "tpm", "coverage"},
        )

    illegal_conditions = [
        (level == "genome_cluster") and (data_type == "gene_abundances"),
        (level == "genomes") and (data_type == "feature_prevalence-ratio"),
        (data_type != "pathway_abundances") and (metric == "coverage"),
    ]
    
    if any(illegal_conditions):
        raise ValueError(f"Invalid combination of arguments: level={level}, data_type={data_type}, metric={metric}")
    
    # Merge tables to produce output
    filepaths = glob.glob(f"{profiling_directory}/*/output/{data_type}.{level}.parquet")
    if filepaths:
        output = dict()
        # Abundance/Coverage
        if data_type in {"feature_abundances", "gene_abundances", "pathway_abundances"}:
            
            # Determine column name
            column = str(metric)
            if data_type in {"feature_abundances", "pathway_abundances"}:
                if metric != "coverage":
                    column = f"{metric}(scaled)"
            
            description = "Merging {}-level {} {} values".format(level, data_type.replace("_", " "), metric)
            for filepath in tqdm(filepaths, description):
                id_sample = filepath.split("/")[-3]
                # df = pd.read_csv(filepath, sep="\t", index_col=[0,1])
                df = pd.read_parquet(filepath)
                output[id_sample] = df[column]
                
        # Prevalence
        elif data_type in {"feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio"}:
            description = "Merging {}-level {} values".format(level, data_type.replace("_", " "))
            for filepath in tqdm(filepaths, description):
                id_sample = filepath.split("/")[-3]
                # df = pd.read_csv(filepath, sep="\t", index_col=0)
                df = pd.read_parquet(filepath)
                output[id_sample] = df.stack()
        X = pd.DataFrame(output).T
        
        sparse_dtype = "float"
        missing_value_fill = pd.NA
        if data_type == "feature_prevalence-binary":
            if fillna_with_zeros:
                X = X.fillna(0)
                sparse_dtype = "int"
                missing_value_fill = 0
        else:
            if fillna_with_zeros:
                X = X.fillna(0.0)
                missing_value_fill = 0.0
        if sparse:
            X = X.astype(pd.SparseDtype(sparse_dtype, missing_value_fill))
        return X
                
    else:
        raise FileNotFoundError(f"Could not find any {data_type}.{level}.parquet files in {profiling_directory}")

def merge_pathway_profiling_tables_as_xarray(profiling_directory:str, data_type:str, level="genomes", metric="number_of_reads", fillna_with_zeros:bool=False):
    
    """
    merges sample-level {data_type} values from multiple samples into a single DataFrame.

    Parameters
    ----------
    profiling_directory : str
        Path to directory containing sample-level directories with output files.
    data_type : str
        Type of {level}-level data to merge. One of: {"feature_abundances", "feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio", "pathway_abundances"}
    level : str, optional
        Level of organization for {data_type}. One of {"genomes", "genome_cluster"}.
    metric : str, optional
        Metric to use for {data_type}. One of {"number_of_reads", "tpm", "coverage"}.
    fillna_with_zeros : bool, optional
        Whether to fill missing values with zeros. Default is False.

    Returns
    -------
    xr.DataArray
        Merged xr.DataArray with dims: (samples, {level}, features)

    Notes
    -----
    Will raise a ValueError if an invalid combination of arguments is provided, such as level="genomes" and data_type="feature_prevalence-ratio".
    
    Files:
    * feature_abundances.genome_clusters.parquet
    * feature_abundances.genomes.parquet
    * feature_prevalence-binary.genome_clusters.parquet
    * feature_prevalence-binary.genomes.parquet
    * feature_prevalence.genome_clusters.parquet
    * feature_prevalence.genomes.parquet
    * feature_prevalence-ratio.genome_clusters.parquet
    * gene_abundances.genomes.parquet
    * pathway_abundances.genome_clusters.parquet
    * pathway_abundances.genomes.parquet
    
    # Not supported: 
    * gene_abundances.genomes.parquet

    """

    check_argument_choice(
        query=data_type, 
        choices={"feature_abundances", "feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio",  "pathway_abundances"}, # "gene_abundances",
        )
    check_argument_choice(
        query=level, 
        choices={"genomes", "genome_clusters"},
        )
    check_argument_choice(
        query=metric, 
        choices={"number_of_reads", "tpm", "coverage"},
        )

    illegal_conditions = [
        # (level == "genome_cluster") and (data_type == "gene_abundances"),
        (level == "genomes") and (data_type == "feature_prevalence-ratio"),
        (data_type != "pathway_abundances") and (metric == "coverage"),
    ]
    
    if any(illegal_conditions):
        raise ValueError(f"Invalid combination of arguments: level={level}, data_type={data_type}, metric={metric}")
    
    # Merge tables to produce output
    filepaths = glob.glob(f"{profiling_directory}/*/output/{data_type}.{level}.parquet")
    if filepaths:
        output = dict()
        # Abundance/Coverage
        if data_type in {"feature_abundances",  "pathway_abundances"}:
            name = (data_type, level, metric)
            variable_label = data_type.split("_")[0] + "s"

            # Determine column name
            column = str(metric)
            if data_type in {"feature_abundances", "pathway_abundances"}:
                if metric != "coverage":
                    column = f"{metric}(scaled)"
            
            description = "Merging {}-level {} {} values".format(level, data_type.replace("_", " "), metric)
            for filepath in tqdm(filepaths, description):
                id_sample = filepath.split("/")[-3]
                # df = pd.read_csv(filepath, sep="\t", index_col=[0,1])
                df = pd.read_parquet(filepath)
                df = df[column].unstack()

                output[id_sample] = xr.DataArray(data = df.values, coords = [(level, df.index), (variable_label, df.columns)])
                del df
                
        # Prevalence
        elif data_type in {"feature_prevalence", "feature_prevalence-binary", "feature_prevalence-ratio"}:
            name = (data_type, level)
            description = "Merging {}-level {} values".format(level, data_type.replace("_", " "))
            for filepath in tqdm(filepaths, description):
                id_sample = filepath.split("/")[-3]
                # df = pd.read_csv(filepath, sep="\t", index_col=0)
                df = pd.read_parquet(filepath)
                variable_label = data_type.split("_")[0] + "s"
                output[id_sample] = xr.DataArray(data = df.values, coords = [(level, df.index), (variable_label, df.columns)])
        X = xr.concat(output.values(), dim="samples")
        X["samples"] = list(output.keys())
        
        
        if data_type in {"feature_prevalence-binary", "feature_prevalence"}:
            X = X.astype(np.int8)
            if fillna_with_zeros:
                X = X.fillna(0)
        else:
            X = X.astype(np.float32)
            if fillna_with_zeros:
                X = X.fillna(0.0)
        return X
                
    else:
        raise FileNotFoundError(f"Could not find any {data_type}.{level}.parquet files in {profiling_directory}")

